"""lg-rez / blocs / Outils pour tables de données

Modification, récupération d'informations sur la structure de la table...

"""

import datetime
import re

import difflib
import unidecode
from sqlalchemy.orm.attributes import flag_modified     # Permet de "signaler" les entrées modifiées, à commit en base

def _remove_accents(s):
    """Renvoie la chaîne non accentuée, mais conserve les caractères spéciaux (emojis...)
    """
    p = re.compile("([À-ʲΆ-ת])")      # Abracadabrax, c'est moche mais ça marche
    return p.sub(lambda c: unidecode.unidecode(c.group()), s)


def modif(item, col, value):
    """Utilitaire : fait ``<item>.<col> = <value>`` et le marque (:func:`~sqlalchemy.orm.attributes.flag_modified`) pour le commit

    Args:
        item (:class:`.bdd.Base`): entrée de BDD à modifier
        col (:class:`str`): colonne à modifier (doit être un attribut valide de la table)
        value (``type(`item`.`col`)``): nouvelle valeur
    """
    setattr(item, col, value)
    flag_modified(item, col)


def transtype(value, col, SQL_type, nullable):
    """Utilitaire : type un input brut selon le type de sa colonne

    Args:
        value (:class:`object`):        valeur à transtyper
        col (:class:`str`):             nom de la colonne associée
        SQL_type (:class:`str`):        type SQL associé, (tel que retourné par :func:`.get_SQL_types` avec ``detail=False``). Types pris en charge : ``"String"``, ``"Integer"``, ``"BigInteger"``, ``"Boolean"``, ``"Time"``, ``"DateTime"``
        nullable (:class:`bool`):       ``True`` si la value peut être ``None`` (cf :func:`.get_SQL_nullable`), ``False`` sinon

    Returns:
        L'objet Python correspondant au type de la colonne (:class:`str`, :class:`int`, :class:`bool`, :class:`datetime.time`, :class:`datetime.datetime`) ou ``None`` (si autorisé)

    Raises:
        ``ValueError``: si la conversion n'est pas possible (ou si ``value`` vaut ``None`` et ``nullable`` est ``False``)
        ``KeyError``: pour un type de colonne non pris en compte
    """
    try:
        if value in (None, '', 'None', 'none', 'Null', 'null', 'not set', 'non défini'):
            if nullable:
                return None
            else:
                raise ValueError
        elif SQL_type == "String":
            return str(value)
        elif SQL_type in ("Integer", "BigInteger"):
            return int(value)
        elif SQL_type == "Boolean":
            if value in [True, 1] or (isinstance(value, str) and value.lower() in ['true', 'vrai', 'on', 'oui', 'yes']):
                return True
            elif value in [False, 0] or (isinstance(value, str) and value.lower() in ['false', 'faux', 'off', 'non', 'no']):
                return False
            else:
                raise ValueError()
        elif SQL_type == "Time":                # hh:mm
            try:
                h, m, _ = value.split(':')
            except ValueError:
                h, m = value.split(':')
            return datetime.time(hour=int(h), minute=int(m))
        elif SQL_type == "DateTime":            # aaaa-mm-jjThh:mm
            date, time = value.split('T')
            aaaa, mm, jj = date.split('-')
            h, m = time.split(':')
            return datetime.datetime(year=int(aaaa), month=int(mm), day=int(jj), hour=int(h), minute=int(m))
        else:
            raise KeyError(f"unknown column type for column '{col}': '{SQL_type}'")
    except (ValueError, TypeError):
        raise ValueError(f"Valeur '{value}' incorrecte pour la colonne '{col}' (type '{SQL_type}'/{'NOT NULL' if not nullable else ''})")


def get_cols(table):
    """Renvoie la liste des noms des colonnes d'une table

    Args:
        table (:class:`.bdd.Base` subclass): table de données

    Returns:
        :class:`list`\[:class:`str`\]
    """
    raw_cols = table.__table__.columns
    return [col.key for col in raw_cols]


def get_primary_col(table):
    """Renvoie le nom de la colonne clé primaire de <table>

    Args:
        table (:class:`.bdd.Base` subclass): table de données

    Returns:
        :class:`str`
    """
    raw_cols = table.__table__.columns
    return [col.key for col in raw_cols if col.primary_key][0]


def get_SQL_types(table, detail=False):
    """Renvoie un dictionnaire {colonne: type SQL}

    Args:
        table (:class:`.bdd.Base` subclass): table de données
        detail (:class:`bool`):

            - Si ``True``, renvoie l'objet type SQL (``col.type``, :class:`sqlalchemy.sql.sqltypes`) : ``String(length=32)``, ``BigInteger``...
            - Si ``False``, renvoie le nom du type (``col.type.__name__``) : ``"String"``, ``"BigInteger"``...

    Returns:
        - :class:`dict`\[:class:`str`, :class:`str`\] (``detail`` ``False``)

    Returns:
        - :class:`dict`\[:class:`str`, :class:`type`\] (``detail`` ``True``)
    """
    raw_cols = table.__table__.columns
    if detail:
        return {col.key: col.type for col in raw_cols}
    else:
        return {col.key: type(col.type).__name__ for col in raw_cols}


def get_SQL_nullable(table):
    """Renvoie un dictionnaire {colonne: accepte les NULL ?}

    Args:
        table (:class:`.bdd.Base` subclass): table de données

    Returns:
        - :class:`dict`\[:class:`str`, :class:`bool`\]
    """
    raw_cols = table.__table__.columns
    return {col.key: col.nullable for col in raw_cols}


def find_nearest(chaine, table, sensi=0.25, filtre=None, carac=None, solo_si_parfait=True, match_first_word=False):
    """Recherche le(s) plus proche résultat(s) dans une table

    Args:
        chaine (:class:`str`): motif à rechercher
        table (:class:`.bdd.Base` subclass): table de données dans laquelle rechercher
        sensi (:class:`float`): ratio minimal pour retenir une entrée
        filtre (:class:`sqlalchemy.sql.elements.BinaryExpression`): argument de :meth:`~sqlalchemy.orm.query.Query.filter` (``Table.colonne == valeur``)
        carac (:class:`str`): colonne selon laquelle rechercher (défaut : colonne primaire de la table)
        solo_si_parfait (:class:`bool`): si ``True``, renvoie uniquement le premier élément de score ``1`` trouvé s'il existe (ignore les autres éléments, même si ``>= sensi``)
        match_first_word (:class:`bool`): si ``True``, teste aussi ``chaine`` vis à vis du premier *mot* (caractères précédent la première espace) de chaque entrée, et conserve ce score si il est supérieur (``table.carac`` doit être de type ``String``).

    Returns:
        :class:`list`\[\(:class:`.bdd.Base`, :class:`float`\)\]: La/les entrée(s) correspondant le mieux à ``chaine``, sous forme de liste de tuples ``(element, score*)`` triés par score\* décroissant

    \*Score = ratio de :class:`difflib.SequenceMatcher`, i.e. proportion de caractères communs aux deux chaînes
    """
    SM = difflib.SequenceMatcher()                      # Création du comparateur de chaînes
    slug1 = _remove_accents(chaine).lower()             # Cible en minuscule et sans accents
    SM.set_seq1(slug1)                                  # Première chaîne à comparer : cible demandée

    if not filtre:
        query = table.query.all()
    else:
        query = table.query.filter(filtre).all()

    scores = []
    if not carac:
        carac = get_primary_col(table)

    for entry in query:
        slug2 = _remove_accents(getattr(entry, carac)).lower()

        SM.set_seq2(slug2)                              # Pour chaque élément, on compare la cible à son nom (en non accentué)
        score = SM.ratio()                              # On calcule la ressemblance

        if match_first_word:
            SM.set_seq2(slug2.split(maxsplit=1)[0])
            scorep = SM.ratio()
            if scorep > score:
                score = scorep

        if score == 1:          # Cas particulier : élément demandé correspondant exactement à un en BDD
            if solo_si_parfait:             # Si demandé, on le renvoie direct
                return [(entry, score)]
            else:                           # Sinon, on ne renverra que les éléments de score 1
                sensi = 1

        scores.append((entry, score))


    bests = [(entry, score) for (entry, score) in sorted(scores, key=lambda x: x[1], reverse=True) if score >= sensi]
    # Meilleurs résultats, dans l'ordre ; si élément parfait trouvé, sensi = 1 donc ne renvoie que les parfaits
    return bests



"""lg-rez / blocs / Émulation de terminal Python (avancé)

Ce module n'a pas grand chose à voir avec le projet dans son ensemble, et était plus une expérimentation personnelle qu'autre chose. Il est néanmoins implémenté par la commande :meth:`\!shell <.bot.Special.Special.shell.callback>`.

Globalement, il s'agit d'une version (très incomplète) du module :mod:`code` de Python, ajoutant un fonctionnement dans un environnement asynchrone. Il y aurait moyen de le réécrire beaucoup plus simplement (et beaucoup plus puissamment) en enveloppant le code à exécuter dans une fonction *async*, en l'envoyant dans un :class:`~code.InteractiveInterpreter` et en *awaitant* le résultat.

Par conséquent, ce module n'est pas considéré comme part de l'API publique ``lgrez`` et non documenté ici. Les plus curieux iront voir le code source !
"""

import re
import sys
import enum
import traceback
import asyncio

import asyncode

from lgrez.blocs import tools


def _py_bloc(data):
    return f"```py\n{data}```"


class RealShellExit(RuntimeError):
    pass

class RSCommand(enum.Enum):
    """Unicode emojis used for shell controll commands"""
    HOME = "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}"
    UNTAB = "\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}"
    TAB = "\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}"
    ENTER = "\N{LEFTWARDS ARROW WITH HOOK}"     # close to "enter" sign
    EOF = "\N{BLACK SQUARE FOR STOP}"


class RealShell(asyncode.AsyncInteractiveConsole):          # Les attributs de cette classe peuvent être modifiés via exec
    """Terminal Python"""

    def __init__(self, bot, channel, locals={}, filename="!shell"):
        """Initialize self"""
        self.bot = bot
        self.channel = channel

        self.message = None
        self.content = ""
        self.indent = 0
        self.rsbuffer = ""

        locals["_shell"] = self
        super().__init__(locals, filename)


    async def interact(self, banner=None, exitmsg="Byebye!"):
        self.message = await self.channel.send(_py_bloc("Launching Python shell...\n"))

        if banner is None:
            cprt = 'Type "help", "copyright", "credits" or "license" for more information.'
            aw = 'async REPL: use "await" directly instead of "asyncio.run()".\n'
            dsc = 'Discord implementation: the caret (‸) indicates the current position of the cursor.\nUse control reactions or send "<", ">", "=" to (un)indent it and send empty lines.'
            banner = f"Python {sys.version} on {sys.platform}\n{cprt}\n{aw}\n{dsc}\n"

        try:
            await super().interact(banner, exitmsg)
        except SystemExit as exc:
            await self.write(f"\n{exitmsg}", refresh=True, show_caret=False)
            raise RealShellExit(*exc.args)
        finally:
            await self.message.clear_reactions()



    async def control(self, command, N=1):
        """Applique l'effet demandé au clic sur un emoji de contrôle du shell"""
        if not isinstance(command, RSCommand):
            command = RSCommand(command)

        if command is RSCommand.HOME and self.indent > 0:       # Reset indentation
            self.indent = 0
            await self.write("", refresh=True)          # on actualise le message
        if command is RSCommand.UNTAB and self.indent > 0:      # Unindent
            self.indent -= max(N, self.indent)
            await self.write("", refresh=True)
        elif command is RSCommand.TAB:          # Indent
            self.indent += N
            await self.write("", refresh=True)
        elif command is RSCommand.ENTER:        # Newline
            return ""       # Send empty str (impossible through Discord)
        elif command is RSCommand.EOF:          # STOP
            raise RealShellExit("Arrêt du shell.")

        return None


    async def get_order(self):
        """Récupère une instruction (message ou contrôle) de l'utilisateur"""
        reacts = self.message.reactions     # Réactons déjà présentes
        if not reacts:
            for command in RSCommand:       # On ajoute les emojis si pas déjà présents (nouveau message)
                await self.message.add_reaction(command.value)
        else:
            try:                                # Sinon, on regarde si un emoji est déjà cliqué
                react = next(r for r in reacts if r.count > 1)
            except StopIteration:                       # Pas le cas ==> on poursuit
                pass
            else:
                async for user in react.users():        # C'est le cas : on trouve l'utilisateur,
                    if user != self.bot.user:
                        await react.remove(user)            # On enlève la réaction,
                        break
                command = RSCommand(react.emoji.name)
                return (await self.control(command))             # Et on applique la commande

        def react_check(payload):       # Check REACT : bon message et pas react du bot
            return (payload.message_id == self.message.id and payload.user_id != self.bot.user.id)

        def message_check(mess):        # Check MESSAGE : bon channel et pas du bot
            return (mess.channel == self.channel and mess.author != self.bot.user)

        react_task = asyncio.create_task(self.bot.wait_for('raw_reaction_add', check=react_check), name="react")
        mess_task = asyncio.create_task(tools.wait_for_message(self.bot, check=message_check, trigger_on_commands=True))
        # On exécute les deux tâche concurremment
        done, pending = await asyncio.wait([react_task, mess_task], return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        done_task = next(iter(done))        # done = tâche complétée

        if done_task.get_name() == "react":     # Clic sur emoji de contrôle
            payload = done_task.result()            # Si clic sur react, done.result = RawReactionActionEvent payload
            await self.message.remove_reaction(payload.emoji, payload.member)
            command = RSCommand(payload.emoji.name)
            return (await self.control(command))

        else:                                   # Réponse par message / STOP
            message = done_task.result()            # Si envoi de message, done.result = Message
            text = message.content
            await message.delete()

            if m := re.match("^<+", text):
                await self.control(RSCommand.UNTAB, N=m.end())
                text = text[m.end():]
            elif m := re.match("^>+", text):
                await self.control(RSCommand.TAB, N=m.end())
                text = text[m.end():]

            if text == "=":
                return (await self.control(RSCommand.ENTER))
            else:
                return (text or None)



    async def write(self, data, *, refresh=False, show_caret=True):
        """Fonction appellée par le shell à chaque print() / repr / traceback...

        N'affiche text que si il se termine par un prompt (sys.ps1 ou sys.ps2) ; l'ajoute à un buffer interne dans le cas contraire.
        Si refresh vaut True, actualise le message quelque soit text.
        """
        if refresh or data.endswith(sys.ps1) or data.endswith(sys.ps2):     # Prompt / refresh demandé
            data = self.rsbuffer + data       # On vide le buffer
            self.rsbuffer = ""
        else:
            self.rsbuffer += data             # On remplit le buffer
            return

        # Auto-indent à l'entrée dans un nouveau bloc
        if data.endswith(f":\n{sys.ps2}"):
            self.indent += 1

        # Ajout caret d'insertion pour indiquer la position du curseur
        if show_caret:
            insert = " "*(self.indent*4) + "‸"
        else:
            insert = ""

        # Actualisation message / nouvau si plein
        if len(self.content + data + insert) < 1990:
            self.content += data
            await self.message.edit(content=_py_bloc(self.content + insert))
        else:
            await self.message.clear_reactions()      # On enlève les commandes du message

            blocs = tools.smooth_split(data, N=1990)
            self.content = blocs[-1]                 # On modifie le texte actuel
            blocs[-1] += insert

            for bloc in blocs:
                message = await self.channel.send(_py_bloc(bloc))

            self.message = message        # On modifie le message actuel = dernier envoyé





    async def raw_input(self, prompt=""):
        """Fonction appellée par le shell à chaque input() / attente d'instruction"""
        if prompt:
            await self.write(prompt)

        entree = None
        while entree is None:
            entree = await self.get_order()      # Modifie l'indentation, stop ou récupère une ligne

        entree = " "*(self.indent*4) + entree
        await self.write(entree + "\n", refresh=True, show_caret=False)       # Affiche l'entrée
        return entree



"""lg-rez / Blocs transversaux

Chaque module de ``lgrez.blocs`` réalise une tâche non liée à une fonctionnalité spécifique du bot : connection à un service, outils divers...

"""

__all__ = ["bdd_tools", "bdd", "env", "gsheets", "pseudoshell", "tools", "webhook"]



"""lg-rez / features / Commandes annexes

Commandes diverses qu'on ne savait pas où ranger

"""

import random
import requests
import traceback
import datetime
import asyncio

from discord.ext import commands
from akinator.async_aki import Akinator
import akinator

from lgrez.blocs import tools
from lgrez.blocs.bdd import session, Joueurs


class Annexe(commands.Cog):
    """Annexe - Commandes annexes aux usages divers"""

    @commands.command()
    async def roll(self, ctx, *, XdY):
        """Lance un ou plusieurs dés

        Args:
            XdY: dés à lancer + modifieurs, au format ``XdY + XdY + ... + Z - Z ...`` avec X le nombre de dés, Y le nombre de faces et Z les modifieurs (constants).

        Examples:
            - ``!roll 1d6``           -> lance un dé à 6 faces
            - ``!roll 1d20 +3``       -> lance un dé à 20 faces, ajoute 3 au résultat
            - ``!roll 1d20 + 2d6 -8`` -> lance un dé 20 plus deux dés 6, enlève 8 au résultat
        """
        dices = XdY.replace(' ', '').replace('-', '+-').split('+')        # "1d6 + 5 - 2" -> ["1d6", "5", "-2"]
        r = ""
        s = 0
        try:
            for dice in dices:
                if 'd' in dice:
                    nb, faces = dice.split('d', maxsplit=1)
                    for i in range(int(nb)):
                        v = random.randrange(int(faces)) + 1
                        s += v
                        r += f" + {v}₍{tools.sub_chiffre(int(faces), True)}₎"
                else:
                    v = int(dice)
                    s += v
                    r += f" {'-' if v < 0 else '+'} {abs(v)}"
            r += f" = {tools.emoji_chiffre(s, True)}"
        except Exception:
            await ctx.send(f"Pattern non reconu. Utilisez {tools.code('!help roll')} pour plus d'informations.")
        else:
            await tools.send_blocs(ctx, r[3:])


    @commands.command(aliases=["cf", "pf"])
    async def coinflip(self, ctx):
        """Renvoie le résultat d'un tirage à Pile ou Face (aléatoire)

        Pile je gagne, face tu perds.
        """
        await ctx.send(random.choice(["Pile", "Face"]))


    @commands.command(aliases=["pong"])
    async def ping(self, ctx):
        """Envoie un ping au bot

        Pong
        """
        ts_rec = datetime.datetime.utcnow()
        delta_rec = ts_rec - ctx.message.created_at     # Temps de réception = temps entre création message et sa réception
        pingpong = ctx.invoked_with.replace("i", "x").replace("I", "X").replace("o", "i").replace("O", "I").replace("x", "o").replace("X", "O")
        cont = (
            f" Réception : {delta_rec.total_seconds()*1000:4.0f} ms\n"
            f" Latence :   {ctx.bot.latency*1000:4.0f} ms\n"
        )
        mess = await ctx.send(f"!{pingpong}\n" + tools.code_bloc(cont + " (...)"))

        ts_ret = datetime.datetime.utcnow()
        delta_ret = ts_ret - mess.created_at            # Retour information message réponse créé
        delta_env = ts_ret - ts_rec - delta_ret         # Temps d'envoi = temps entre réception 1er message (traitement quasi instantané) et création 2e, moins temps de retour d'information
        delta_tot = delta_rec + delta_ret + delta_env   # Total = temps entre création message !pong et réception information réponse envoyée
        await mess.edit(content=f"!{pingpong}\n" + tools.code_bloc(cont +
            f" Envoi :     {delta_env.total_seconds()*1000:4.0f} ms\n"
            f" Retour :    {delta_ret.total_seconds()*1000:4.0f} ms\n"
            f"——————————————————————\n"
            f" Total :     {delta_tot.total_seconds()*1000:4.0f} ms"
        ))


    @commands.command()
    @tools.mjs_only
    async def addhere(self, ctx, *joueurs):
        """Ajoute les membres au chan courant (COMMANDE MJ)

        Args:
            *joueurs: membres à ajouter, chacun entouré par des guillemets si nom + prénom

        Si ``*joueurs`` est un seul élément, il peut être de la forme ``<crit>=<filtre>`` tel que décrit dans l'aide de ``!send``.
        """
        ts_debut = ctx.message.created_at - datetime.timedelta(microseconds=1)

        if len(joueurs) == 1 and "=" in joueurs[0]:      # Si critère : on remplace joueurs
            crit, filtre = joueurs[0].split("=", maxsplit=1)
            if hasattr(Joueurs, crit):
                joueurs = Joueurs.query.filter(getattr(Joueurs, crit) == filtre).all()
            else:
                await ctx.send(f"Critère \"{crit}\" incorrect. !help {ctx.invoked_with} pour plus d'infos.")
                return
        else:                                           # Sinon, si noms / mentions
            joueurs = [await tools.boucle_query_joueur(ctx, cible) for cible in joueurs]

        for joueur in joueurs:
            member = ctx.guild.get_member(joueur.discord_id)
            assert member, f"Member {joueur} introuvable"
            await ctx.channel.set_permissions(member, read_messages=True, send_messages=True)
            await ctx.send(f"{joueur.nom} ajouté")

        mess = await ctx.send("Fini, purge les messages ?")
        if await tools.yes_no(ctx.bot, mess):
            await ctx.channel.purge(after=ts_debut)


    @commands.command()
    @tools.mjs_only
    async def purge(self, ctx, N=None):
        """Supprime tous les messages de ce chan (COMMANDE MJ)

        Args:
            N: nombre de messages à supprimer (défaut : tous)
        """
        if N:
            mess = await ctx.send(f"Supprimer les {N} messages les plus récents de ce chan ? (celui-ci inclus)")
        else:
            mess = await ctx.send(f"Supprimer tous les messages de ce chan ?")

        if await tools.yes_no(ctx.bot, mess):
            await ctx.channel.purge(limit=int(N) if N else None)


    @commands.command()
    async def akinator(self, ctx):
        """J'ai glissé chef

        Implémentation directe de https://pypi.org/project/akinator.py
        """
        # Un jour mettre ça dans des embeds avec les https://fr.akinator.com/bundles/elokencesite/images/akitudes_670x1096/<akitude>.png croppées, <akitude> in ["defi", "serein", "inspiration_legere", "inspiration_forte", "confiant", "mobile", "leger_decouragement", "vrai_decouragement", "deception", "triomphe"]
        await ctx.send(f"Vous avez demandé à être mis en relation avec {tools.ital('Akinator : Le Génie du web')}.\nVeuillez patienter...")
        async with ctx.typing():
            # Connection
            aki = Akinator()
            question = await aki.start_game(language="fr")

        exit = False
        while not exit and aki.progression <= 80:
            mess = await ctx.send(f"({aki.step + 1}) {question}")
            reponse = await tools.wait_for_react_clic(ctx.bot, mess, {"👍":"yes", "🤷":"idk", "👎":"no", "⏭️":"stop"})
            if reponse == "stop":
                exit = True
            else:
                async with ctx.typing():
                    question = await aki.answer(reponse)

        async with ctx.typing():
            await aki.win()

        mess = await ctx.send(f"Tu penses à {tools.bold(aki.first_guess['name'])} ({tools.ital(aki.first_guess['description'])}) !\nJ'ai bon ?\n{aki.first_guess['absolute_picture_path']}")
        if await tools.yes_no(ctx.bot, mess):
            await ctx.send("Yay\nhttps://fr.akinator.com/bundles/elokencesite/images/akitudes_670x1096/triomphe.png")
        else:
            await ctx.send("Oof\nhttps://fr.akinator.com/bundles/elokencesite/images/akitudes_670x1096/deception.png")


    @commands.command()
    async def xkcd(self, ctx, N):
        """J'ai aussi glissé chef, mais un peu moins

        Args:
            N: numéro du comic
        """
        async with ctx.typing():
            r = requests.get(f"https://xkcd.com/{N}/info.0.json")

        if not r:
            await ctx.send("Paramètre incorrect ou service non accessible.")
            return

        url = r.json().get("img")
        if not url:
            await ctx.send("Paramètre incorrect ou service non accessible.")
            return

        await ctx.send(url)



"""lg-rez / features / Actions publiques

Gestion des haros, candidatures à la mairie, résultats des votes

"""

import discord
from discord.ext import commands

from lgrez.blocs import bdd, tools
from lgrez.blocs.bdd import Joueurs, CandidHaro


class ActionsPubliques(commands.Cog):
    """ActionsPubliques - Commandes pour gérer les actions vous engageant publiquement"""

    @commands.command()
    @tools.vivants_only
    @tools.private
    async def haro(self, ctx, *, cible=None):
        """Lance publiquement un haro contre un autre joueur

        Args:
            cible: nom du joueur à accuser

        Cette commande n'est utilisable que lorsqu'un vote pour le condamné est en cours.
        """
        auteur = ctx.author

        joueur = Joueurs.query.get(auteur.id)
        assert joueur, f"Joueur {auteur.display_name} introuvable"

        if joueur.vote_condamne_ is None:
            await ctx.send("Pas de vote pour le condamné de jour en cours !")
            return

        cible = await tools.boucle_query_joueur(ctx, cible, 'Contre qui souhaites-tu déverser ta haine ?')

        if cible.statut == "mort":
            await ctx.send("Nan mais oh, tu crois qu'il a pas assez souffert en mourant lui ?")
            return

        elif cible.statut == "immortel":
            await ctx.send(f"Comment oses-tu t'en prendre à ceux qui te sont supérieurs ? {tools.role(ctx, 'MJ').mention}, regardez un peu ce qu'il se passe là...")
            return

        await tools.send_blocs(ctx, "Et quelle est la raison de cette haine, d'ailleurs ?")
        motif = await tools.wait_for_message_here(ctx)

        emb = discord.Embed(title = f"**{tools.emoji(ctx, 'ha')}{tools.emoji(ctx, 'ro')} contre {cible.nom} !**",
                            description = f"**« {motif.content} »\n**",
                            color=0xff0000)
        emb.set_author(name=f"{ctx.author.display_name} en a gros 😡😡")
        emb.set_thumbnail(url=tools.emoji(ctx, "bucher").url)
        emb.set_footer(text=f"Utilise !vote {cible.nom} pour voter contre lui.")

        mess = await ctx.send("C'est tout bon ?", embed=emb)
        if await tools.yes_no(ctx.bot, mess):
            cible_member = ctx.guild.get_member(cible.discord_id)
            assert cible_member, f"!haro : Member {cible} introuvable"

            if not CandidHaro.query.filter_by(player_id=cible.discord_id, type="haro").all():       # Inscription haroté
                haroted = CandidHaro(player_id=cible.discord_id, type="haro")
                bdd.session.add(haroted)

            if not CandidHaro.query.filter_by(player_id=ctx.author.id, type="haro").all():          # Inscription haroteur
                haroteur = CandidHaro(player_id=ctx.author.id, type="haro")
                bdd.session.add(haroteur)

            bdd.session.commit()

            await tools.channel(ctx, "haros").send(f"(Psst, {cible_member.mention} :3)", embed=emb)
            await tools.channel(ctx, "débats").send(f"{tools.emoji(ctx, 'ha')}{tools.emoji(ctx, 'ro')} de {auteur.mention} sur {cible_member.mention} ! Vous en pensez quoi vous ? (détails sur {tools.channel(ctx, 'haros').mention})")
            await ctx.send(f"Allez, c'est parti ! ({tools.channel(ctx, 'haros').mention})")

        else:
            await ctx.send("Mission aborted.")


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def candid(self, ctx):
        """Candidate à l'élection du nouveau maire

        Cette commande n'est utilisable que lorsqu'un vote pour le nouveau maire est en cours.
        """
        joueur = Joueurs.query.get(ctx.author.id)
        assert joueur, f"Joueur {auteur.display_name} introuvable"

        if joueur.vote_maire_ is None:
            await ctx.send("Pas de vote pour le nouveau maire en cours !")
            return

        if CandidHaro.query.filter_by(player_id=ctx.author.id, type = "candidature").all():
            await ctx.send("Hola collègue, tout doux, tu t'es déjà présenté !")
            return

        auteur = ctx.author

        await tools.send_blocs(ctx, "Quel est ton programme politique ?")
        motif = await tools.wait_for_message_here(ctx)

        emb = discord.Embed(title = f"**{tools.emoji(ctx, 'maire')} {auteur.display_name} candidate à la Mairie !** {tools.emoji(ctx, 'mc', must_be_found=False) or ''}",
                            description = "Voici ce qu'il a à vous dire :\n" + tools.bold(motif.content),
                            color=0xf1c40f)
        emb.set_author(name=f"{auteur.display_name} vous a compris !")
        emb.set_thumbnail(url=tools.emoji(ctx, "maire").url)
        emb.set_footer(text=f"Utilise !votemaire {auteur.display_name} pour voter pour lui.")

        mess = await ctx.send("C'est tout bon ?", embed=emb)
        if await tools.yes_no(ctx.bot, mess):
            candidat = CandidHaro(id=None, player_id=ctx.author.id, type="candidature")
            bdd.session.add(candidat)
            bdd.session.commit()

            await tools.channel(ctx, "haros").send("Here comes a new challenger !", embed=emb)
            await tools.channel(ctx, "débats").send(f"{auteur.mention} se présente à la Mairie ! Vous en pensez quoi vous ?\n (détails sur {tools.channel(ctx, 'haros').mention})")
            await ctx.send(f"Allez, c'est parti ! ({tools.channel(ctx, 'haros').mention})")


    ##Fonctions de gestion de la base CandidHaro (wipe les haros ou les votes)
    @commands.command()
    @tools.mjs_only
    async def wipe(self, ctx, quoi):
        """Efface les haros / candidatures du jour (COMMANDE MJ)

        Args:
            quoi (str): peut être

                - ``haros`` : Supprimer les haros
                - ``candids`` : Supprimer les candicatures
        """
        if quoi == "haros":
            items = CandidHaro.query.filter_by(type="haro").all()
        elif quoi == "candids":
            items = CandidHaro.query.filter_by(type="candidature").all()
        else:
            await ctx.send("Mauvais argument")

        if not items:
            await ctx.send("Rien à faire")
            await tools.log(ctx, f"!wipe {quoi} : rien à faire")
        else:
            for item in items:
                bdd.session.delete(item)
            bdd.session.commit()
            await ctx.send("Fait.")
            await tools.log(ctx, f"!wipe {quoi} : fait")


    @commands.command(enabled=False)
    async def listharo(self, ctx):
        """Liste les gens qui ont subi un haro aujourd'hui"""
        mess = "Les gens que tu pourras (peut être) voir sur le bûcher aujourd'hui sont:\n"
        haroted = CandidHaro.query.filter_by(type="haro").all()
        if not haroted:
            await ctx.send("Le village est encore calme, personne n'a encore accusé personne...")
        else:
            for joueur in haroted:
                mess += f"- {Joueurs.query.filter_by(discord_id = joueur.player_id).first().nom} \n"
        await tools.send_code_blocs(ctx, mess)


    @commands.command(enabled=False)
    async def listcandid(self, ctx):
        """Liste les candidats à la mairie aujourd'hui"""
        mess = "Les candidats à la mairie pour aujourd'hui sont les suivants :\n"
        candids = CandidHaro.query.filter_by(type="candidature").all()
        if not candids:
            await ctx.send("Pas de chance, personne ne s'est présenté...")
            return
        else:
            for joueur in candids:
                mess += f"- {Joueurs.query.filter_by(discord_id = joueur.player_id).first().nom} \n"
        await tools.send_code_blocs(ctx, mess)



"""lg-rez / features / Gestion des votes et actions

Ouverture / fermeture / rappels des votes et actions (+ refill)

"""

import datetime

from discord.ext import commands
from sqlalchemy.sql.expression import and_, or_, not_

from lgrez.blocs import tools, bdd, bdd_tools
from lgrez.features import gestion_actions, taches
from lgrez.blocs.bdd import Joueurs, Actions, BaseActions, BaseActionsRoles, CandidHaro


async def recup_joueurs(quoi, qui, heure=None):
    """Renvoie les joueurs concernés par la tâche !quoi <qui> [heure]

    Args:
        quoi (:class:`str`): évènement, ``"open" / "close" / "remind"``
        qui (:class:`str`): cible, ``"cond" / "maire" / "loups" / "action"``
        heure (:class:`str`): si ``qui == "action"``, heure associée (au format ``HHhMM``)

    Returns:
        :class:`list`\[:class:`.bdd.Joueurs`\]

    Examples:
        ``!open cond`` -> joueurs avec droit de vote
        ``!close action 17h`` -> joueurs dont l'action se termine à 17h
    """
    criteres = {
        "cond": {
            "open": and_(Joueurs.votant_village == True,        # Objets spéciaux SQLAlchemy.BinaryExpression : ne PAS simplifier !!!
                         Joueurs.vote_condamne_ == None),
            "close": Joueurs.vote_condamne_ != None,
            "remind": Joueurs.vote_condamne_ == "non défini",
            },
        "maire": {
            "open": and_(Joueurs.votant_village == True,        # Objets spéciaux SQLAlchemy.BinaryExpression : ne PAS simplifier !!!
                         Joueurs.vote_maire_ == None),
            "close": Joueurs.vote_maire_ != None,
            "remind": Joueurs.vote_maire_ == "non défini",
            },
        "loups": {
            "open": and_(Joueurs.votant_loups == True,          # Objets spéciaux SQLAlchemy.BinaryExpression : ne PAS simplifier !!!
                         Joueurs.vote_loups_ == None),
            "close": Joueurs.vote_loups_ != None,
            "remind": Joueurs.vote_loups_ == "non défini",
            },
        }

    if qui in criteres:
        critere = criteres[qui][quoi]
        return Joueurs.query.filter(critere).all()      # Liste des joueurs répondant aux critères

    elif qui == "action":
        if heure and isinstance(heure, str):            # Si l'heure est précisée, on convertit str "HHhMM" -> datetime.time
            tps = tools.heure_to_time(heure)
        else:                                           # Si l'heure n'est pas précisée, on prend l'heure actuelle
            raise ValueError("[heure] doit être spécifiée lorque <qui> == \"action\"")
            # tps = datetime.datetime.now().time()
            # if quoi == "remind":
            #     tps += datetime.timedelta(hours=1)      # Si remind, on considère l'heure qui arrive

        actions = await gestion_actions.get_actions(quoi, "temporel", tps)

        dic = {}
        for action in actions:
            if (joueur := Joueurs.query.get(action.player_id)) in dic:
                dic[joueur].append(action)
            else:
                dic[joueur] = [action]

        return dic
        #Formerly :
        #{joueur.player_id:[action for action in actions if action.player_id == joueur.player_id] for joueur in [Joueurs.query.get(action.player_id) for action in actions]}

    elif qui.isdigit() and (action := Actions.query.get(int(qui))):     # Appel direct action par son numéro
        if ((quoi == "open" and (not action.decision_       # Sécurité : ne pas lancer une action déjà lancée,
                    or action.trigger_debut == "perma"))        # (sauf si permanente ==> ré-ouverture)
            or (quoi == "close" and action.decision_)       #   ni fermer une déjà fermée
            or (quoi == "remind" and action.decision_ == "rien")):

            return {Joueurs.query.get(action.player_id): [action]}
        else:
            return {}

    else:
        raise ValueError(f"""Argument <qui> == \"{qui}" invalide""")



class OpenClose(commands.Cog):
    """OpenClose - Commandes de lancement, rappel et fermeture des votes et actions"""

    @commands.command()
    @tools.mjs_only
    async def open(self, ctx, qui, heure=None, heure_chain=None):
        """Lance un vote / des actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions commençant à ``heure``
                ``{id}``        pour une action spécifique (paramètre :attr:`.bdd.Actions.id`)
                ===========     ===========

            heure:
                - si ``qui == "cond"``, ``"maire"`` ou ``"loup"``, programme en plus la fermeture à ``heure`` (et un rappel 30 minutes avant) ;
                - si ``qui == "action"``, il est obligatoire : heure des actions à lancer (cf plus haut). Pour les actions, la fermeture est de toute façon programmée le cas échéant (``trigger_fin`` ``temporel`` ou ``delta``).

                Dans tous les cas, format ``HHh`` ou ``HHhMM``.

            heure_chain:
                permet de chaîner des votes : lance le vote immédiatement et programme sa fermeture à ``heure``, en appellant ``!close`` de sorte à programmer une nouvelle ouverture le lendemain à ``heure_chain``, et ainsi de suite.
                Format ``HHh`` ou ``HHhMM``.

        Une sécurité empêche de lancer un vote ou une action déjà en cours.

        Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer / planifier la fermeture des votes, par exemple).

        Examples:
            - ``!open maire``             lance un vote condamné maintenant
            - ``!open cond 19h``          lance un vote condamné maintenant et programme sa fermeture à 19h00 (ex. Juge Bègue)
            - ``!open cond 18h 10h``      lance un vote condamné maintenant, programme sa fermeture à 18h00, et une prochaine ouverture à 10h, etc
            - ``!open action 19h``        lance toutes les actions commençant à 19h00
            - ``!open 122``               lance l'action d'ID 122

        """
        joueurs = await recup_joueurs("open", qui, heure)        # Liste de joueurs (votes) ou dictionnaire joueur : action

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await tools.send_code_blocs(ctx, f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n - {str_joueurs}")

        for joueur in joueurs:
            chan = ctx.guild.get_channel(joueur.chan_id_)
            assert chan, f"!open : chan privé de {joueur} introuvable"

            if qui == "cond":
                bdd_tools.modif(joueur, "vote_condamne_", "non défini")
                message = await chan.send(
                    f"""{tools.montre()}  Le vote pour le condamné du jour est ouvert !  {tools.emoji(ctx, "bucher")} \n"""
                    + (f"""Tu as jusqu'à {heure} pour voter. \n""" if heure else "")
                    + tools.ital(f"""Tape {tools.code('!vote (nom du joueur)')} ou utilise la réaction pour voter."""))
                await message.add_reaction(tools.emoji(ctx, "bucher"))

            elif qui == "maire":
                bdd_tools.modif(joueur, "vote_maire_", "non défini")
                message = await chan.send(
                    f"""{tools.montre()}  Le vote pour l'élection du maire est ouvert !  {tools.emoji(ctx, "maire")} \n"""
                    + (f"""Tu as jusqu'à {heure} pour voter. \n""" if heure else "")
                    + tools.ital(f"""Tape {tools.code('!votemaire (nom du joueur)')} ou utilise la réaction pour voter."""))
                await message.add_reaction(tools.emoji(ctx, "maire"))

            elif qui == "loups":
                bdd_tools.modif(joueur, "vote_loups_", "non défini")
                message = await chan.send(
                    f"""{tools.montre()}  Le vote pour la victime de cette nuit est ouvert !  {tools.emoji(ctx, "lune")} \n"""
                    + (f"""Tu as jusqu'à {heure} pour voter. \n""" if heure else "")
                    + tools.ital(f"""Tape {tools.code('!voteloups (nom du joueur)')} ou utilise la réaction pour voter."""))
                await message.add_reaction(tools.emoji(ctx, "lune"))

            else:       # Action
                for action in joueurs[joueur]:
                    await gestion_actions.open_action(ctx, action, chan)

        bdd.session.commit()

        # Actions déclenchées par ouverture
        for action in Actions.query.filter_by(trigger_debut=f"open_{qui}"):
            await gestion_actions.open_action(ctx, action)

        # Réinitialise haros/candids
        items = []
        if qui == "cond":
            items = CandidHaro.query.filter_by(type="haro").all()
        elif qui == "maire":
            items = CandidHaro.query.filter_by(type="candidature").all()
        if items:
            for item in items:
                bdd.session.delete(item)
            bdd.session.commit()
            await tools.log(ctx, f"!open {qui} : haros/candids wiped")
            await tools.channel(ctx, "haros").send(f"{tools.emoji(ctx, 'void')}\n"*30
                + "Nouveau vote, nouveaux haros !\n"
                + tools.ital(f"Les posts ci-dessus sont invalides pour le vote actuel. Utilisez {tools.code('!haro')} pour en relancer."))

        # Programme fermeture
        if qui in ["cond", "maire", "loups"] and heure:
            ts = tools.next_occurence(tools.heure_to_time(heure))
            taches.add_task(ctx.bot, ts - datetime.timedelta(minutes=30), f"!remind {qui}")
            if heure_chain:
                taches.add_task(ctx.bot, ts, f"!close {qui} {heure_chain} {heure}")      # Programmera prochaine ouverture
            else:
                taches.add_task(ctx.bot, ts, f"!close {qui}")



    @commands.command()
    @tools.mjs_only
    async def close(self, ctx, qui, heure=None, heure_chain=None):
        """Ferme un vote / des actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions se terminant à ``heure``
                ``{id}``        pour une action spécifique (paramètre :attr:`.bdd.Actions.id`)
                ===========     ===========

            heure:
                - si ``qui == "cond"``, ``"maire"`` ou ``"loup"``, programme en plus une prochaine ouverture à ``heure`` (et un rappel 30 minutes avant) ;
                - si ``qui == "action"``, il est obligatoire : heure des actions à lancer (cf plus haut). Pour les actions, la prochaine est de toute façon programmée le cas échéant (cooldown à 0 et reste des charges).

                Dans tous les cas, format ``HHh`` ou ``HHhMM``.

            heure_chain:
                permet de chaîner des votes : ferme le vote immédiatement et programme une prochaine ouverture à ``heure``, en appellant ``!close`` de sorte à programmer une nouvelle fermeture le lendemain à ``heure_chain``, et ainsi de suite.
                Format ``HHh`` ou ``HHhMM``.

        Une sécurité empêche de fermer un vote ou une action qui n'est pas en cours.

        Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi (penser à envoyer / planifier la fermeture des votes, par exemple).

        Examples:
            - ``!close maire``            ferme le vote condamné maintenant
            - ``!close cond 10h``         ferme le vote condamné maintenant et programme une prochaine ouverture à 10h00
            - ``!close cond 10h 18h``     ferme le vote condamné maintenant, programme une prochaine ouverture à 10h00, qui sera fermé à 18h, etc
            - ``!close action 22h``       ferme toutes les actions se terminant à 22h00
            - ``!close 122``              ferme l'action d'ID 122
        """

        joueurs = await recup_joueurs("close", qui, heure)

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await ctx.send(tools.code_bloc(f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n{str_joueurs}"))

        for joueur in joueurs:
            chan = ctx.guild.get_channel(joueur.chan_id_)
            assert chan, f"!close : chan privé de {joueur} introuvable"

            if qui == "cond":
                await chan.send(f"""{tools.montre()}  Fin du vote pour le condamné du jour ! \n"""
                                f"""Vote définitif : {joueur.vote_condamne_}\n"""
                                f"""Les résultats arrivent dans l'heure !\n""")
                bdd_tools.modif(joueur, "vote_condamne_", None)

            elif qui == "maire":
                await chan.send(f"""{tools.montre()}  Fin du vote pour le maire ! \n"""
                                f"""Vote définitif : {joueur.vote_maire_}""")
                bdd_tools.modif(joueur, "vote_maire_", None)

            elif qui == "loups":
                await chan.send(f"""{tools.montre()}  Fin du vote pour la victime du soir ! \n"""
                                f"""Vote définitif : {joueur.vote_loups_}""")
                bdd_tools.modif(joueur, "vote_loups_", None)

            else:       # Action
                for action in joueurs[joueur]:
                    await chan.send(f"""{tools.montre()}  Fin de la possiblité d'utiliser ton action {tools.code(action.action)} ! \n"""
                                    f"""Action définitive : {action.decision_}""")
                    await gestion_actions.close_action(ctx, action, chan)

        bdd.session.commit()

        # Actions déclenchées par fermeture
        for action in Actions.query.filter_by(trigger_debut=f"close_{qui}"):
            await gestion_actions.open_action(ctx, action)

        # Programme prochaine ouverture
        if qui in ["cond", "maire", "loups"] and heure:
            ts = tools.next_occurence(tools.heure_to_time(heure))
            if heure_chain:
                taches.add_task(ctx.bot, ts, f"!open {qui} {heure_chain} {heure}")      # Programmera fermeture
            else:
                taches.add_task(ctx.bot, ts, f"!open {qui}")



    @commands.command()
    @tools.mjs_only
    async def remind(self, ctx, qui, heure=None):
        """Envoi un rappel de vote / actions de rôle (COMMANDE BOT / MJ)

        Args:
            qui:
                ===========     ===========
                ``cond``        pour le vote du condamné
                ``maire``       pour le vote du maire
                ``loups``       pour le vote des loups
                ``action``      pour les actions se terminant à ``heure``
                ``{id}``        pour une action spécifique (paramètre :attr:`.bdd.Actions.id`)
                ===========     ===========

            heure: ne sert que dans le cas où <qui> == "action" (il est alors obligatoire), contrairement à !open et !close.
                Format HHh ou HHhMM.

        Le bot n'envoie un message qu'aux joueurs n'ayant pas encore voté / agi, si le vote ou l'action est bien en cours.

        Cette commande a pour vocation première d'être exécutée automatiquement par des tâches planifiées.
        Elle peut être utilisée à la main, mais attention à ne pas faire n'importe quoi !.

        Examples:
            - ``!remind maire``           rappelle le vote condamné maintenant
            - ``!remind action 22h``      rappelle toutes les actions se terminant à 22h00
            - ``!remind 122``             rappelle l'action d'ID 122
        """

        joueurs = await recup_joueurs("remind", qui, heure)

        str_joueurs = "\n - ".join([joueur.nom for joueur in joueurs])
        await ctx.send(tools.code_bloc(f"Utilisateur(s) répondant aux critères ({len(joueurs)}) : \n{str_joueurs}"))

        for joueur in joueurs:
            chan = ctx.guild.get_channel(joueur.chan_id_)
            assert chan, f"!remind : chan privé de {joueur} introuvable"
            member = ctx.guild.get_member(joueur.discord_id)
            assert member, f"!remind : member {joueur} introuvable"

            if qui == "cond":
                message = await chan.send(f"""⏰ {member.mention} Plus que 30 minutes pour voter pour le condamné du jour ! 😱 \n""")
                await message.add_reaction(tools.emoji(ctx, "bucher"))

            elif qui == "maire":
                message = await chan.send(f"""⏰ {member.mention} Plus que 30 minutes pour élire le nouveau maire ! 😱 \n""")
                await message.add_reaction(tools.emoji(ctx, "maire"))

            elif qui == "loups":
                message = await chan.send(f"""⏰ {member.mention} Plus que 30 minutes voter pour la victime du soir ! 😱 \n""")
                await message.add_reaction(tools.emoji(ctx, "lune"))

            else:       # Action
                for action in joueurs[joueur]:
                    message = await chan.send(f"""⏰ {member.mention} Plus que 30 minutes pour utiliser ton action {tools.code(action.action)} ! 😱 \n""")
                    await message.add_reaction(tools.emoji(ctx, "action"))



    @commands.command()
    @tools.mjs_only
    async def refill(self, ctx, motif, *, cible=None):
        """Permet de recharger le/les pouvoirs rechargeables (COMMANDE BOT / MJ)

        Args:
            motif: ``"weekends"``, ``"forgeron"``, ``"rebouteux"`` ou ``"divin"`` (forcer le refill car les MJs tout-puissants l'ont décidé)
            cible: ``"all"`` ou le nom d'un joueur
        """
        if motif not in ["weekends", "forgeron", "rebouteux", "divin"]:
            await ctx.send(f"{motif} n'est pas un <motif> valide")
            return

        if motif == "divin":
            if cible != "all":
                target = await tools.boucle_query_joueur(ctx, cible=cible, message="Qui veux-tu recharger ?")
                refillable = Actions.query.filter(Actions.charges != None).filter_by(player_id=target.discord_id).all()
            else:
                m = await ctx.send("Tu as choisi de recharger le pouvoir de TOUS les joueurs actifs, en es-tu sûr ?")

                if await tools.yes_no(ctx.bot, m):
                    refillable = Actions.query.filter(Actions.charges != None).all()

                else:
                    await ctx.send("Mission aborted.")
                    return

        else: #refill WE, forgeron ou rebouteux
            if cible != "all":
                target = await tools.boucle_query_joueur(ctx, cible=cible, message = "Qui veux-tu recharger ?")
                refillable = Actions.query.filter(Actions.refill.contains(motif)).filter_by(player_id=target.discord_id).all()
            else:
                refillable = Actions.query.filter(Actions.refill.contains(motif)).all()

        await tools.log(ctx, refillable)

        txt = "Action(s) répondant aux critères :\n"
        for action in refillable:
            txt += f"- {action.action}, id = {action.id} \n"

        await tools.send_code_blocs(ctx, txt)

        # Détermination nouveau nombre de charges
        if motif == "weekends":
            remplissage = {action: BaseActions.query.get(action.action).base_charges for action in refillable}
        else:
            remplissage = {action: action.charges + 1 for action in refillable}

        # Refill proprement dit
        for action, charge in remplissage.items():
            if charge > action.charges:
                if not action.charges and action.trigger_debut == "perma":           # Action permanente : on ré-ouvre !
                    ts = tools.fin_pause() if motif == "weekends" else datetime.datetime.now() + datetime.timedelta(seconds=10)
                    taches.add_task(ctx.bot, ts, f"!open {action.id}", action=action.id)

                bdd_tools.modif(action, "charges", charge)

                await tools.send_blocs(tools.private_chan(ctx.guild.get_member(action.player_id)), f"Ton action {action.action} vient d'être rechargée, tu as maintenant {charge} charge(s) disponible(s) !")

        bdd.session.commit()


    @commands.command()
    @tools.mjs_only
    async def cparti(self, ctx):
        """Lance le jeu (COMMANDE MJ)

        - Crée (et programme) les actions associées aux rôles de tous les joueurs     ==> EN FAIT NON, plus besoin vu que c'est fait à la synchro des rôles
        - Programme les votes condamnés quotidiens (avec chaînage) 10h-18h
        - Programme un vote maire 10h-18h
        - Programme les actions au lancement du jeu (choix de mentor...) et permanentes (forgeron)... à 19h

        À utiliser le jour du lancement après 10h (lance les premières actions le soir et les votes le lendemain)
        """

        message = await ctx.send("C'est parti ?\nLes rôles ont bien été attribués et synchronisés ? (si non, le faire AVANT de valider)\n\nOn est bien après 10h le jour du lancement ?\n\nTu es conscient que tous les joueurs reçevront à 18h55 un message en mode « happy Hunger Games » ? (codé en dur parce que flemme)")
        if await tools.yes_no(ctx.bot, message):
            async with ctx.typing():
                joueurs = Joueurs.query.all()
                r = "C'est parti !\n"


                r += "\n\nBon plus besoin de lancer les actions, c'est fait à la synchro des rôles mais les !open n'ont aucun impact tant que tout le monde est en role_actif == False, DU COUP passer tout le monde en True genre MAINTENANT (et en silencieux !) si on veut vraiment lancer\n\n"

                # Programmation votes condamnés chainés 10h-18h
                r += "\nProgrammation des votes :\n"
                taches.add_task(ctx.bot, tools.next_occurence(datetime.time(hour=10)), "!open cond 18h 10h")
                r += " - À 10h : !open cond 18h 10h\n"

                # Programmation votes loups chainés 19h-23h
                taches.add_task(ctx.bot, tools.next_occurence(datetime.time(hour=19)), "!open loups 23h 19h")
                r += " - À 19h : !open loups 23h 19h\n"

                # Programmation premier vote maire 10h-17h
                taches.add_task(ctx.bot, tools.next_occurence(datetime.time(hour=10)), "!open maire 17h")
                r += " - À 10h : !open maire 17h\n"

                # Programmation actions au lancement et actions permanentes
                r += "\nProgrammation des actions start / perma :\n"
                ts = tools.next_occurence(datetime.time(hour=19))
                for action in Actions.query.filter_by(trigger_debut="start").all() + Actions.query.filter_by(trigger_debut="perma").all():
                    r += f" - À 19h : !open {action.id} (trigger_debut == {action.trigger_debut})\n"
                    taches.add_task(ctx.bot, ts, f"!open {action.id}", action=action.id)

                # Programmation refill weekends
                # r += "\nProgrammation des refills weekends :\n"
                # taches.add_task(ctx.bot, tools.fin_pause() - datetime.timedelta(minutes=5), f"!refill weekends all")
                # r += " - Dimanche à 18h55 : !refill weekends all\n"

                # Programmation envoi d'un message aux connards
                r += f"\nEt, à 18h50 : !send all [message de hype oue oue c'est génial]\n"
                taches.add_task(ctx.bot, ts - datetime.timedelta(minutes=10), "!send all Ah {member.mention}... J'espère que tu es prêt(e), parce que la partie commence DANS 10 MINUTES !!! https://tenor.com/view/thehungergames-hungergames-thggifs-effie-gif-5114734")

                await tools.log(ctx, r, code=True)

            await ctx.send("C'est tout bon ! (normalement) (détails dans #logs)")
        else:
            await ctx.send("Mission aborted.")



"""lg-rez / features / Commandes informatives

Commandes donnant aux joueurs des informations sur le jeu, leurs actions, les joueurs en vie et morts...

"""

import traceback
import unidecode

from discord.ext import commands

from lgrez.blocs import bdd, bdd_tools, tools
from lgrez.blocs.bdd import session, Joueurs, Roles, Actions, BaseActions


class Informations(commands.Cog):
    """Informations - Commandes disponibles pour en savoir plus sur soi et les autres"""

    @commands.command(aliases=["role", "rôles", "rôle", "camp", "camps"])
    async def roles(self, ctx, *, filtre=None):
        """Affiche la liste des rôles / des informations sur un rôle

        Args:
            filtre: peut être

                - "Villageois", "Loups", "Nécros", "Autres" pour les rôles d'un camp ;
                - Un nom de rôle pour les informations sur ce rôle.

        Sans argument liste tous les rôles existants.
        """
        if filtre:
            filtre = tools.remove_accents(filtre.lower())

        if not filtre:
            roles = Roles.query.order_by(Roles.nom).all()
        elif filtre in ("villageois", "village", "<villageois>", "<village>"):
            roles = Roles.query.filter_by(camp="village").all()
        elif filtre in ("loups", "<loups>"):
            roles = Roles.query.filter_by(camp="loups").all()
        elif filtre in ("necros", "<necros>"):
            roles = Roles.query.filter_by(camp="nécro").all()
        elif filtre in ("autres", "solitaires", "<autres>", "<solitaires>"):
            roles = Roles.query.filter_by(camp="solitaire").all()
        elif filtre:
            if role := Roles.query.get(filtre):     # Slug du rôle trouvé direct
                pass
            else:                                   # Sinon, on cherche en base
                roles = bdd_tools.find_nearest(filtre, Roles, carac="nom")
                if roles:
                    role = roles[0][0]
                else:
                    await ctx.send(f"Rôle / camp \"{filtre}\" non trouvé.")
                    return
            await ctx.send(tools.code_bloc(f"{role.prefixe}{role.nom} – {role.description_courte} (camp : {role.camp})\n\n{role.description_longue}"))
            return

        await tools.send_blocs(ctx,
            f"Rôles trouvés :\n"
            + "\n".join([str(tools.emoji_camp(ctx, role.camp)) + tools.code(f"{role.nom.ljust(25)} {role.description_courte}") for role in roles if not role.nom.startswith("(")])
            + "\n" + tools.ital(f"({tools.code('!roles <role>')} pour plus d'informations sur un rôle.)")
        )


    @commands.command()
    @tools.mjs_only
    async def rolede(self, ctx, *, cible):
        """Donne le rôle d'un joueur (COMMANDE MJ)

        Args:
            cible: le joueur dont on veut connaître le rôle
        """
        joueur = await tools.boucle_query_joueur(ctx, cible, "Qui ?")
        if joueur:
            await ctx.send(tools.nom_role(joueur.role, prefixe=True))


    @commands.command()
    @tools.mjs_only
    async def quiest(self, ctx, *, nomrole):
        """Liste les joueurs ayant un rôle donné (COMMANDE MJ)

        Args:
            nomrole: le rôle qu'on cherche (doit être un slug ou nom de rôle valide)
        """
        roles = bdd_tools.find_nearest(tools.remove_accents(nomrole), Roles, carac="slug")
        if roles:
            role = roles[0][0]
        else:
            roles = await bdd_tools.find_nearest(tools.remove_accents(nomrole).capitalize(), Roles, carac="nom")
            if roles:
                role = roles[0][0]
            else:
                await ctx.send("Connais pas")

        joueurs = Joueurs.query.filter_by(role=role.slug).filter(Joueurs.statut.in_(["vivant", "MV"])).all()
        await ctx.send(f"{tools.nom_role(role.slug)} : " + (", ".join(joueur.nom for joueur in joueurs) if joueurs else "Personne."))


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def menu(self, ctx):
        """Affiche des informations et boutons sur les votes / actions en cours

        Le menu a une place beaucoup moins importante ici que sur Messenger, vu que tout est accessible par commandes.
        """
        member = ctx.author
        joueur = Joueurs.query.get(member.id)
        assert joueur, f"!menu : joueur {member} introuvable"

        reacts = []
        r = "––– MENU –––\n\n"

        if joueur.vote_condamne_:
            r += f" - {tools.emoji(ctx, 'bucher')}  Vote pour le bûcher en cours – vote actuel : {tools.code(joueur.vote_condamne_)}\n"
            reacts.append(tools.emoji(ctx, 'bucher'))
        if joueur.vote_maire_:
            r += f" - {tools.emoji(ctx, 'maire')}  Vote pour le maire en cours – vote actuel : {tools.code(joueur.vote_maire_)}\n"
            reacts.append(tools.emoji(ctx, 'maire'))
        if joueur.vote_loups_:
            r += f" - {tools.emoji(ctx, 'lune')}  Vote des loups en cours – vote actuel : {tools.code(joueur.vote_loups_)}\n"
            reacts.append(tools.emoji(ctx, 'lune'))

        if not reacts:
            r += "Aucun vote en cours.\n"

        actions = Actions.query.filter_by(player_id=member.id).filter(Actions.decision_ != None).all()
        if actions:
            for action in actions:
                r += f" - {tools.emoji(ctx, 'action')}  Action en cours : {tools.code(action.action)} (id {action.id}) – décision : {tools.code(action.decision_)}\n"
            reacts.append(tools.emoji(ctx, 'action'))
        else:
            r += "Aucune action en cours.\n"

        message = await ctx.send(r + f"\n{tools.code('!infos')} pour voir ton rôle et tes actions, {tools.code('@MJ')} en cas de problème")
        for react in reacts:
            await message.add_reaction(react)


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def infos(self, ctx):
        """Affiche tes informations de rôle / actions

        Toutes les actions liées à ton rôle (et parfois d'autres) sont indiquées, même celles que tu ne peux pas utiliser pour l'instant (plus de charges, déclenchées automatiquement...)
        """
        def disponible(action):
            """Renvoie une description human readible des conditions de déclenchement d'<action>"""
            if action.trigger_debut == "temporel":
                dispo = f"De {tools.time_to_heure(action.heure_debut)} à {tools.time_to_heure(action.heure_fin)}"
            elif action.trigger_debut == "perma":
                dispo = f"N'importe quand"
            elif action.trigger_debut == "start":
                dispo = f"Au lancement de la partie"
            elif action.trigger_debut == "mort":
                dispo = f"S'active à ta mort"
            elif action.trigger_debut == "mot_mjs":
                dispo = f"S'active à l'annonce des résultats du vote"
            elif "_" in action.trigger_debut:
                quoi, qui = action.trigger_debut.split("_")
                d_quoi = {"open": "l'ouverture",
                          "close": "la fermeture",
                          "remind": "personne utilise ça",
                }
                d_qui = {"cond": "du vote condamné",
                         "maire": "du vote pour la mairie",
                         "loups": "du vote pour les loups",
                         "action": "personne utilise ça",
                }
                try:
                    dispo = f"S'active à {d_quoi[quoi]} {d_qui[qui]}"
                except KeyError:
                    dispo = f"S'active à {quoi}_{qui}"
            else:
                dispo = action.trigger_debut

            return dispo


        member = ctx.author
        joueur = Joueurs.query.get(member.id)
        assert joueur, f"!infos : joueur {member} introuvable"
        r = ""

        r += f"Ton rôle actuel : {tools.bold(tools.nom_role(joueur.role) or joueur.role)}\n"
        r += tools.ital(f"({tools.code(f'!roles {joueur.role}')} pour tout savoir sur ce rôle)")

        actions = Actions.query.filter_by(player_id=member.id).all()

        if actions:
            r += "\n\nActions :"
            r += tools.code_bloc("\n".join([(
                f" - {str(action.action).ljust(20)} "
                + (f"Cooldown : {cooldown}" if (cooldown := action.cooldown) else disponible(action)).ljust(22)
                + (f"   {action.charges} charge(s){' pour cette semaine' if (action.refill and 'weekends' in action.refill) else ''}" if isinstance(action.charges, int) else "Illimitée")     # Vraiment désolé pour cette immondice
                ) for action in actions]
            ))
        else:
            r += "\n\nAucune action disponible."

        await ctx.send(r + f"\n{tools.code('!menu')} pour voir les votes et actions en cours, {tools.code('@MJ')} en cas de problème")


    @commands.command()
    @tools.mjs_only
    async def actions(self, ctx, *, cible):
        """Affiche et modifie les actions d'un joueur (COMMANDE MJ)

        Args:
            cible: le joueur dont on veut voir ou modifier les actions
        """
        joueur = await tools.boucle_query_joueur(ctx, cible, "Qui ?")
        if joueur:

            member = ctx.guild.get_member(joueur.discord_id)
            assert member, f"!actions : membre {joueur} introuvable"
            r = ""

            r += f"Rôle : {tools.nom_role(joueur.role) or joueur.role}\n"

            actions = Actions.query.filter_by(player_id=joueur.discord_id).all()

            if actions:
                r += "Actions :"
                r += tools.code_bloc(
                    f"#️⃣  id   action                   début     fin       cd   charges   refill\n"
                    "----------------------------------------------------------------------------------\n"
                    + "\n".join([(tools.emoji_chiffre(i+1) + "  "
                        + str(action.id).ljust(5)
                        + str(action.action).ljust(25)
                        + str(action.heure_debut if action.trigger_debut == "temporel" else action.trigger_debut).ljust(10)
                        + str(action.heure_fin if action.trigger_fin == "temporel" else action.trigger_fin).ljust(10)
                        + str(action.cooldown).ljust(5)
                        + str(action.charges).ljust(10)
                        + str(action.refill)
                    ) for i, action in enumerate(actions)]
                ))
                r += "Modifier/ajouter/stop :"
                message = await ctx.send(r)
                i = await tools.choice(ctx.bot, message, len(actions), additionnal={"⏺": -1, "⏹": 0})
                if i > 0:       # modifier
                    action = actions[i-1]
                    stop = False
                    while not stop:
                        await ctx.send(f"Modifier : (parmis {tools.code('début, fin, cd, charges, refill')})\n{tools.code('valider')} pour finir\n(Utiliser {tools.code('!open id')}/{tools.code('!close id')} pour ouvrir/fermer)")
                        modif = (await tools.wait_for_message_here(ctx)).content.lower()

                        if modif in ["début", "fin"]:
                            await ctx.send(f"Trigger (parmis {tools.code('temporel, delta, perma, start, auto, mort, mot_mjs, {open|close|remind}_{cond|maire|loups}')}) ou heure direct si {tools.code('temporel')} :")
                            trigger = (await tools.wait_for_message_here(ctx)).content.lower()
                            heure = None
                            if ":" in trigger or "h" in trigger:
                                heure = trigger
                                trigger = "temporel"

                            if trigger in ["temporel", "delta"]:
                                if not heure:
                                    await ctx.send(f"Heure / delta {tools.code('HHhMM ou HH:MM')} :")
                                    heure = (await tools.wait_for_message_here(ctx)).content
                                ts = tools.heure_to_time(heure)
                                bdd_tools.modif(action, "trigger_debut" if modif == "début" else "trigger_fin", trigger)
                                bdd_tools.modif(action, "heure_debut" if modif == "début" else "heure_fin", ts)
                            elif trigger in ["perma", "start", "auto", "mot_mjs", "mort"] + [f"{quoi}_{qui}" for quoi in ["open", "close", "remind"] for qui in ["cond", "maire", "loups"]]:
                                bdd_tools.modif(action, "trigger_debut" if modif == "début" else "trigger_fin", trigger)
                                bdd_tools.modif(action, "heure_debut" if modif == "début" else "heure_fin", None)
                            else:
                                await ctx.send("Valeur incorrecte")

                        elif modif in ["cd", "cooldown"]:
                            await ctx.send(f"Combien ?")
                            cd = int((await tools.wait_for_message_here(ctx)).content)
                            bdd_tools.modif(action, "cooldown", cd)

                        elif modif == "charges":
                            await ctx.send(f"Combien ? ({tools.code('None')} pour illimité)")
                            entry = (await tools.wait_for_message_here(ctx)).content
                            charges = None if entry.lower() == "none" else int(entry)
                            bdd_tools.modif(action, "charges", charges)

                        elif modif == "refill":
                            await ctx.send(f"Quoi ? ({tools.code('rebouteux / forgeron / weekends')} séparés par des {tools.code(', ')})")
                            refill = (await tools.wait_for_message_here(ctx)).content.lower()
                            bdd_tools.modif(action, "refill", refill)

                        elif modif == "valider":
                            bdd.session.commit()
                            await ctx.send("Fait.")

                        else:
                            await ctx.send("Valeur incorrecte")

                elif i < 0:     # ajouter
                    await ctx.send("Pas encore codé, pas de chance")

            else:
                r += "Aucune action pour ce joueur."
                await ctx.send(r)



    @commands.command(aliases=["joueurs", "vivant"])
    async def vivants(self, ctx):
        """Affiche la liste des joueurs vivants

        Aussi dite : « liste des joueurs qui seront bientôt morts »
        """
        joueurs = [joueur for joueur in Joueurs.query.filter(Joueurs.statut != "mort").order_by(Joueurs.nom)]

        mess = f" Joueur                     en chambre\n"
        mess += f"––––––––––––––––––––––––––––––––––––––––––––––\n"
        if ctx.bot.config.get("demande_chambre", True):
            for joueur in joueurs:
                mess += f" {joueur.nom.ljust(25)}  {joueur.chambre}\n"
        else:
            for joueur in joueurs:
                mess += f" {joueur.nom}\n"

        await tools.send_code_blocs(ctx, mess, prefixe=f"Les {len(joueurs)} joueurs vivants sont :")


    @commands.command(aliases=["mort"])
    async def morts(self, ctx):
        """Affiche la liste des joueurs morts

        Aussi dite : « liste des joueurs qui mangent leurs morts »
        """
        joueurs = [joueur for joueur in Joueurs.query.filter_by(statut="mort").order_by(Joueurs.nom)]

        if joueurs:
            mess = ""
            for joueur in joueurs:
                mess += f" {joueur.nom}\n"
        else:
            mess = "Toi (mais tu ne le sais pas encore)"

        await tools.send_code_blocs(ctx, mess, prefixe=f"Les {len(joueurs) or ''} morts sont :")



import sys
import asyncio
import logging
import traceback
import time
import datetime
import re

import discord
from discord.ext import commands

from lgrez import __version__
from lgrez.blocs import tools, bdd, env, bdd_tools, gsheets, webhook, realshell
from lgrez.blocs.bdd import Tables, Joueurs, Actions, Roles
from lgrez.features import annexe, IA, inscription, informations, sync, open_close, voter_agir, taches, actions_publiques, communication


logging.basicConfig(level=logging.WARNING)



### Checks et système de blocage

class AlreadyInCommand(commands.CheckFailure):
    """Exception levée lorsq'un member veut lancer une commande dans un salon déjà occupé.

    Classe fille de :exc:`discord.ext.commands.CheckFailure`.
    """
    pass

async def already_in_command(ctx):
    """Check (:class:`discord.ext.commands.Check`) : vérifie si le joueur est actuellement dans une commande.

    Si non, raise une exception :exc:`AlreadyInCommand`.

    Args:
        ctx (:class:`discord.ext.commands.Context`): contexte d'invocation de la commande.
    """
    if (ctx.channel.id in ctx.bot.in_command
        and ctx.command.name not in ["stop", "panik"]):     # Cas particuliers : !stop/!panik

        await ctx.send("stop", delete_after=0)      # On envoie (discrètement) l'ordre d'arrêt commande précédente
        await asyncio.sleep(1)                      # On attend qu'il soit pris en compte
        if ctx.channel.id in ctx.bot.in_command:    # Si ça n'a pas suffit
            raise AlreadyInCommand()                    # on raise l'erreur
        else:
            return True
    else:
        return True

# @bot.before_invoke
async def add_to_in_command(ctx):
    """Ajoute le channel de ``ctx`` à la liste des channels dans une commande.

    Cette fonction est appellée avant chaque appel de fonction.
    Elle est appellée seulement si les checks sont OK, donc pas si le salon est déjà dans :attr:`ctx.bot.in_command <.LGBot.in_command>`.

    Args:
        ctx (:class:`discord.ext.commands.Context`): contexte d'invocation de la commande.
    """
    if ctx.command.name != "stop" and not ctx.message.webhook_id:
        ctx.bot.in_command.append(ctx.channel.id)

# @bot.after_invoke
async def remove_from_in_command(ctx):
    """Retire le channel de ``ctx`` de la liste des channels dans une commande.

    Cette fonction est appellée après chaque appel de fonction.
    Elle attend 0.5 secondes avant d'enlever le joueur afin d'éviter que le bot réagisse « nativement » (IA) à un message déjà traité par un :func:`.tools.wait_for_message` ayant mené à la fin de la commande.

    Args:
        ctx (:class:`discord.ext.commands.Context`): contexte d'invocation de la commande.
    """
    await asyncio.sleep(0.5)        # On attend un peu
    if ctx.channel.id in ctx.bot.in_command:
        ctx.bot.in_command.remove(ctx.channel.id)



### Réactions aux différents évènements

# Au démarrage du bot
async def _on_ready(bot):
    """Méthode appellée par Discord au démarrage du bot.

    Vérifie le serveur, log et affiche publiquement que le bot est fonctionnel ; restaure les tâches planifiées éventuelles et exécute celles manquées.

    Si :attr:`bot.config <.LGBot.config>` ``["output_liveness"]`` vaut ``True``, lance :attr:`bot.i_am_alive <.LGBot.i_am_alive>` (écriture chaque minute sur un fichier disque)
    """
    if bot.config.get("output_liveness"):
        bot.i_am_alive()            # Start liveness regular output

    guild = bot.get_guild(bot.GUILD_ID)
    assert guild, f"on_ready : Guilde d'ID {bot.GUILD_ID} introuvable"

    print(f"{bot.user} connecté au serveur « {guild.name} » (id : {guild.id})\n")
    print(f"Guild Members: " + " - ".join([member.display_name for member in guild.members]))
    print(f"\nChannels: " + " - ".join([channel.name for channel in guild.text_channels]))

    await tools.log(guild, "Juste rebooted!")
    await bot.change_presence(activity=discord.Activity(type=discord.ActivityType.listening, name="vos demandes (!help)"))

    # Tâches planifiées
    now = datetime.datetime.now()
    r = 0

    for tache in Tables["Taches"].query.all():
        delta = (tache.timestamp - now).total_seconds()
        TH = bot.loop.call_later(delta, taches.execute, tache, bot.loop)  # Si delta < 0 (action manquée), l'exécute immédiatement, sinon attend jusqu'à tache.timestamp
        bot.tasks[tache.id] = TH
        # r += f"Récupération de la tâche {tools.code(tache.timestamp.strftime('%d/%m/%Y %H:%M:%S'))} > {tools.code(tache.commande)}\n"
        r += 1

    if r:
        await tools.log(guild, f"{r} tâches planifiées récupérées en base et reprogrammées.")


# À l'arrivée d'un membre sur le serveur
async def _on_member_join(bot, member):
    """Méthode appellée par Discord à l'arrivée d'un joueur sur le serveur.

    Log et lance le processus d'inscription.

    Ne fait rien si l'arrivée n'est pas sur le serveur :attr:`GUILD_ID`.

    Args:
        member (:class:`discord.Member`): Le joueur qui vient d'arriver.
    """
    if member.guild.id != bot.GUILD_ID:            # Bon serveur
        return

    await tools.log(member, f"Arrivée de {member.name}#{member.discriminator} sur le serveur")
    await inscription.main(bot, member)


# Au départ d'un membre du serveur
async def _on_member_remove(bot, member):
    """Méthode appellée par Discord au départ d'un joueur du serveur.

    Log en mentionnant les MJs.

    Ne fait rien si le départ n'est pas du serveur :attr:`GUILD_ID`.

    Args:
        member (:class:`discord.Member`): Le joueur qui vient de partir.
    """
    if member.guild.id != bot.GUILD_ID:            # Bon serveur
        return

    await tools.log(member, f"{tools.mention_MJ(member)} ALERTE : départ de {member.display_name} ({member.name}#{member.discriminator}) du serveur !!")


# À chaque message
async def _on_message(bot, message):
    """Méthode appellée par Discord à la réception d'un message.

    Invoque l'ensemble des commandes, ou les règles d'IA si
        - Le message n'est pas une commande
        - Le message est posté dans un channel privé (#conv-bot-...)
        - Il n'y a pas déjà de commande en cours dans ce channel
        - Le channel n'est pas en mode STFU

    Ne fait rien si le message n'est pas sur le serveur :attr:`GUILD_ID` ou qu'il est envoyé par le bot lui-même ou par un membre sans aucun rôle affecté.

    Args:
        member (:class:`discord.Member`): Le joueur qui vient d'arriver.
    """

    if message.author == bot.user:              # Sécurité pour éviter les boucles infinies
        return

    if not message.guild:                       # Message privé
        await message.channel.send("Je n'accepte pas les messages privés, désolé !")
        return

    if message.guild.id != bot.GUILD_ID:        # Mauvais serveur
        return

    if not message.webhook_id:                  # Pas un webhook
        if message.author.top_role == tools.role(message, "@everyone"):     # Pas de rôle affecté
            return                                                              # le bot te calcule même pas

    if message.content.startswith(bot.command_prefix + " "):
        message.content = bot.command_prefix + message.content[2:]

    await bot.invoke(await bot.get_context(message))        # On trigger toutes les commandes
    # (ne PAS remplacer par bot.process_commands(message), en théorie c'est la même chose mais ça détecte pas les webhooks...)

    if (not message.content.startswith(bot.command_prefix)  # Si pas une commande
        and message.channel.name.startswith("conv-bot")     # et dans un channel de conversation bot
        and message.channel.id not in bot.in_command        # et pas déjà dans une commande (vote...)
        and message.channel.id not in bot.in_stfu):         # et le channel est pas en mode STFU

        await IA.process_IA(bot, message)                       # On trigger les règles d'IA


# À chaque réaction ajoutée
async def _on_raw_reaction_add(bot, payload):
    """Méthode appellée par Discord à l'ajout d'une réaction sur un message.

    Appelle la fonction adéquate si le joueur est en base et a cliqué sur ":bucher:", ":maire:", ":lune:" ou ":action:".

    Ne fait rien si la réaction n'est pas sur le serveur :attr:`GUILD_ID`.

    Args:
        payload (:class:`discord.RawReactionActionEvent`): Paramètre "statique" (car le message n'est pas forcément dans le cache du bot, par exemple si il a été reboot depuis).

    Quelques attributs utiles :
        - ``payload.member`` (:class:`discord.Member`) : Membre ayant posé la réaction
        - ``payload.emoji`` (:class:`discord.PartialEmoji`) : PartialEmoji envoyé
        - ``payload.message_id`` (int) : ID du message réacté
    """
    if payload.guild_id != bot.GUILD_ID:            # Mauvais serveur
        return

    reactor = payload.member
    if reactor == bot.user:
        return

    if payload.emoji == tools.emoji(reactor, "bucher"):
        if not Joueurs.query.get(reactor.id):        # Gens pas en base
            return

        ctx = await tools.create_context(bot, reactor, "!vote")
        if ctx.guild.get_channel(payload.channel_id).name.startswith("conv-bot"):
            await ctx.send(f"""{payload.emoji} > {tools.bold("Vote pour le condamné du jour :")}""")
            await bot.invoke(ctx)       # On trigger !vote

    elif payload.emoji == tools.emoji(reactor, "maire"):
        if not Joueurs.query.get(reactor.id):        # Gens pas en base
            return

        ctx = await tools.create_context(bot, reactor, "!votemaire")
        if ctx.guild.get_channel(payload.channel_id).name.startswith("conv-bot"):
            await ctx.send(f"""{payload.emoji} > {tools.bold("Vote pour le nouveau maire :")}""")
            await bot.invoke(ctx)       # On trigger !votemaire

    elif payload.emoji == tools.emoji(reactor, "lune"):
        if not Joueurs.query.get(reactor.id):        # Gens pas en base
            return

        ctx = await tools.create_context(bot, reactor, "!voteloups")
        if ctx.guild.get_channel(payload.channel_id).name.startswith("conv-bot"):
            await ctx.send(f"""{payload.emoji} > {tools.bold("Vote pour la victime des loups :")}""")
            await bot.invoke(ctx)       # On trigger !voteloups

    elif payload.emoji == tools.emoji(reactor, "action"):
        if not Joueurs.query.get(reactor.id):        # Gens pas en base
            return

        ctx = await tools.create_context(bot, reactor, "!action")
        if ctx.guild.get_channel(payload.channel_id).name.startswith("conv-bot"):
            await ctx.send(f"""{payload.emoji} > {tools.bold("Action :")}""")
            await bot.invoke(ctx)       # On trigger !voteloups


# Gestion des erreurs dans les commandes
async def _on_command_error(bot, ctx, exc):
    """Méthode appellée par Discord à chaque exception levée dans une commande.

    Analyse l'erreur survenue et informe le joueur de manière adéquate en fonction, en mentionnant les MJs si besoin.

    Ne fait rien si l'exception n'a pas eu lieu sur le serveur :attr:`GUILD_ID`.

    Args:
        ctx (:class:`discord.ext.commands.Context`): Contexte dans lequel l'exception a été levée
        exc (:class:`discord.ext.commands.CommandError`): Exception levée
    """
    if ctx.guild.id != bot.GUILD_ID:            # Mauvais serveur
        return

    if isinstance(exc, commands.CommandInvokeError) and isinstance(exc.original, tools.CommandExit):     # STOP envoyé
        await ctx.send(str(exc.original) or "Mission aborted.")

    elif isinstance(exc, commands.CommandInvokeError):
        if isinstance(exc.original, bdd.SQLAlchemyError) or isinstance(exc.original, bdd.DriverOperationalError):       # Erreur SQL
            if bdd.session:
                await tools.log(ctx, "Rollback session")
                bdd.session.rollback()          # On rollback la session

        await ctx.send(f"Oups ! Un problème est survenu à l'exécution de la commande  :grimacing:\n"
                       f"{tools.mention_MJ(ctx)} ALED – "
                       f"{tools.ital(f'{type(exc.original).__name__}: {str(exc.original)}')}")

    elif isinstance(exc, commands.CommandNotFound):
        await ctx.send(f"Hum, je ne connais pas cette commande  :thinking:\n"
                       f"Utilise {tools.code('!help')} pour voir la liste des commandes.")

    elif isinstance(exc, commands.DisabledCommand):
        await ctx.send(f"Cette commande est désactivée. Pas de chance !")

    elif isinstance(exc, commands.ConversionError) or isinstance(exc, commands.UserInputError):
        await ctx.send(f"Hmm, ce n'est pas comme ça qu'on utilise cette commande ! Petit rappel : ({tools.code(f'{type(exc).__name__}: {exc}')})")
        ctx.message.content = f"!help {ctx.command.name}"
        ctx = await bot.get_context(ctx.message)
        await ctx.reinvoke()

    elif isinstance(exc, commands.CheckAnyFailure):         # Normalement raise que par @tools.mjs_only
        await ctx.send("Hé ho toi, cette commande est réservée aux MJs !  :angry:")

    elif isinstance(exc, commands.MissingAnyRole):          # Normalement raise que par @tools.joueurs_only
        await ctx.send("Cette commande est réservée aux joueurs ! (parce qu'ils doivent être inscrits en base, toussa)"
                       f"({tools.code('!doas')} est là en cas de besoin)")

    elif isinstance(exc, commands.MissingRole):             # Normalement raise que par @tools.vivants_only
        await ctx.send("Désolé, cette commande est réservée aux joueurs en vie !")

    elif isinstance(exc, AlreadyInCommand) and ctx.command.name not in ["addIA", "modifIA"]:
        await ctx.send(f"Impossible d'utiliser une commande pendant un processus ! (vote...)\n"
                       f"Envoie {tools.code('stop')} pour arrêter le processus.")

    elif isinstance(exc, commands.CheckFailure):        # Autre check non vérifié
        await ctx.send(f"Tiens, il semblerait que cette commande ne puisse pas être exécutée ! {tools.mention_MJ(ctx)} ?\n"
                       f"({tools.ital(f'{type(exc).__name__}: {str(exc)}')})")

    else:
        await ctx.send(f"Oups ! Une erreur inattendue est survenue  :grimacing:\n"
                       f"{tools.mention_MJ(ctx)} ALED – "
                       f"{tools.ital(f'{type(exc).__name__}: {str(exc)}')}")


# Erreurs non gérées par le code précédent (hors du cadre d'une commande)
async def _on_error(bot, event, *args, **kwargs):
    """Méthode appellée par Discord à chaque exception remontant au-delà d'une commande.

    Log en mentionnant les MJs. Cette méthode permet de gérer les exceptions sans briser la loop du bot (i.e. il reste en ligne).

    Args:
        event (str): Nom de l'évènement ayant généré une erreur (``"member_join"``, ``"message"``...)
        *args, \**kwargs: Arguments passés à la fonction traitant l'évènement : ``member``, ``message``...
    """
    etype, exc, tb = sys.exc_info()         # Exception ayant causé l'appel

    guild = bot.get_guild(bot.GUILD_ID)
    assert guild, f"on_error : Serveur {bot.GUILD_ID} introuvable - Erreur initiale : \n{traceback.format_exc()}"

    if isinstance(exc, bdd.SQLAlchemyError) or isinstance(exc, bdd.DriverOperationalError):     # Erreur SQL
        if bdd.session:
            bdd.session.rollback()          # On rollback la session
            await tools.log(guild, "Rollback session")

    await tools.log(guild, traceback.format_exc(), code=True,
        prefixe=f"{tools.role(guild, 'MJ').mention} ALED (with prefix) : Exception Python !")

    raise        # On remonte l'exception à Python (pour log, ça ne casse pas la loop)



### Commandes spéciales

class Special(commands.Cog):
    """Special - Commandes spéciales (méta-commandes, imitant ou impactant le déroulement des autres)"""

    @commands.command(aliases=["kill"])
    @tools.mjs_only
    async def panik(self, ctx):
        """Tue instantanément le bot, sans confirmation (COMMANDE MJ)

        PAAAAANIK
        """
        sys.exit()


    @commands.command()
    @tools.mjs_only
    async def do(self, ctx, *, code):
        """Exécute du code Python et affiche le résultat (COMMANDE MJ)

        Args:
            code: instructions valides dans le contexte du fichier ``bot.py`` (utilisables notemment : ``ctx``, ``bdd.session``, ``Tables``...)

        Si ``code`` est une coroutine, elle sera awaited (ne pas inclure ``await`` dans ``code``).

        Aussi connue sous le nom de « faille de sécurité », cette commande permet de faire environ tout ce qu'on veut sur le bot (y compris le crasher, importer des modules, exécuter des fichiers .py... même si c'est un peu compliqué) voire d'impacter le serveur sur lequel le bot tourne si on est motivé.

        À utiliser avec parcimonie donc, et QUE pour du développement/debug !
        """
        class Answer():
            def __init__(self):
                self.rep = ""
        a = Answer()
        exec(f"a.rep = {code}")
        if asyncio.iscoroutine(a.rep):
            a.rep = await a.rep
        await ctx.send(f"Entrée : {tools.code(code)}\nSortie :\n{a.rep}")


    @commands.command()
    @tools.mjs_only
    async def shell(self, ctx):
        """Lance un pseudo-terminal Python (COMMANDE MJ)

        Envoyer ``help`` dans le pseudo-terminal pour plus d'informations sur son fonctionnement.

        Évidemment, les avertissements dans ``!do`` s'appliquent ici : ne pas faire n'imp avec cette commande !! (même si ça peut être très utile, genre pour ajouter des gens en masse à un channel)
        """
        locs = globals()
        locs["ctx"] = ctx
        shell = realshell.RealShell(ctx.bot, ctx.channel, locs)
        try:
            await shell.interact()
        except realshell.RealShellExit as exc:
            raise tools.CommandExit(*exc.args if exc else "!shell: Forced to end.")


    @commands.command()
    @tools.mjs_only
    async def co(self, ctx, cible=None):
        """Lance la procédure d'inscription comme si on se connectait au serveur pour la première fois (COMMANDE MJ)

        Args:
            cible: la MENTION (``@joueur``) du joueur à inscrire, par défaut le lançeur de la commande.

        Cette commande est principalement destinée aux tests de développement, mais peut être utile si un joueur chibre son inscription (à utiliser dans son channel, ou ``#bienvenue`` (avec ``!autodestruct``) si même le début a chibré).
        """
        if cible:
            id = ''.join([c for c in cible if c.isdigit()])         # Si la chaîne contient un nombre, on l'extrait
            if id and (member := ctx.guild.get_member(int(id))):          # Si c'est un ID d'un membre du serveur
                pass
            else:
                await ctx.send("Cible introuvable.")
                return
        else:
            member = ctx.author

        await inscription.main(ctx.bot, member)


    @commands.command()
    @tools.mjs_only
    async def doas(self, ctx, *, qui_quoi):
        """Exécute une commande en tant qu'un autre joueur (COMMANDE MJ)

        Args:
            qui_quoi: nom de la cible (nom ou mention d'un joueur INSCRIT) suivi de la commande à exécuter (commençant par un ``!``).

        Example:
            ``!doas Vincent Croquette !vote Annie Colin``
        """
        qui, quoi = qui_quoi.split(" " + ctx.bot.command_prefix, maxsplit=1)       # !doas <@!id> !vote R ==> qui = "<@!id>", quoi = "vote R"
        joueur = await tools.boucle_query_joueur(ctx, qui.strip() or None, Tables["Joueurs"])
        member = ctx.guild.get_member(joueur.discord_id)
        assert member, f"!doas : Member {joueur} introuvable"

        ctx.message.content = ctx.bot.command_prefix + quoi
        ctx.message.author = member

        await ctx.send(f":robot: Exécution en tant que {joueur.nom} :")
        await remove_from_in_command(ctx)       # Bypass la limitation de 1 commande à la fois
        await ctx.bot.process_commands(ctx.message)
        await add_to_in_command(ctx)


    @commands.command(aliases=["autodestruct", "ad"])
    @tools.mjs_only
    async def secret(self, ctx, *, quoi):
        """Supprime le message puis exécute la commande (COMMANDE MJ)

        Args:
            quoi: commande à exécuter, commençant par un ``!``

        Utile notemment pour faire des commandes dans un channel public, pour que la commande (moche) soit immédiatement supprimée.
        """
        await ctx.message.delete()

        ctx.message.content = quoi

        await remove_from_in_command(ctx)       # Bypass la limitation de 1 commande à la fois
        await ctx.bot.process_commands(ctx.message)
        await add_to_in_command(ctx)


    @commands.command()
    @tools.private
    async def stop(self, ctx):
        """Peut débloquer des situations compliquées (beta)

        Ne pas utiliser cette commande sauf en cas de force majeure où plus rien ne marche, et sur demande d'un MJ (après c'est pas dit que ça marche mieux après l'avoir utilisée)
        """
        if ctx.channel.id in ctx.bot.in_command:
            ctx.bot.in_command.remove(ctx.channel.id)
        ctx.send("Te voilà libre, camarade !")


    ### 6 bis - Gestion de l'aide

    @commands.command(aliases=["aide", "aled", "oskour"])
    async def help(self, ctx, *, command=None):
        """Affiche la liste des commandes utilisables et leur utilisation

        Args:
            command (optionnel): nom exact d'une commande à expliquer (ou un de ses alias)

        Si ``command`` n'est pas précisée, liste l'ensemble des commandes accessibles à l'utilisateur.
        """

        pref = ctx.bot.command_prefix
        cogs = ctx.bot.cogs                                                                 # Dictionnaire nom: cog
        commandes = {cmd.name: cmd for cmd in ctx.bot.commands}                             # Dictionnaire nom: commande
        aliases = {alias: nom for nom, cmd in commandes.items() for alias in cmd.aliases}   # Dictionnaire alias: nom de la commande

        len_max = max(len(cmd) for cmd in commandes)

        if not command:         # Pas d'argument ==> liste toutes les commandes
            ctx.bot.in_command.remove(ctx.channel.id)   # On désactive la limitation de une commande simultanée sinon can_run renvoie toujours False
            async def filter_runnables(commands):           # Obligé parce que can_run doit être await, donc c'est compliqué
                """Retourne la liste des commandes pouvant run parmis commands"""
                runnables = []
                for cmd in commands:
                    try:
                        runnable = await cmd.can_run(ctx)
                    except Exception:       # Parfois can_run raise une exception pour dire que la commande est pas runnable
                        runnable = False
                    if runnable:
                        runnables.append(cmd)
                return runnables

            r = f"""{ctx.bot.description} (v{__version__})"""
            for cog in cogs.values():
                runnables = await filter_runnables(cog.get_commands())
                if runnables:                           # pour chaque cog contenant des runnables
                    r += f"\n\n{cog.description} :"
                    for cmd in runnables:               # pour chaque commande runnable
                        r += f"\n  - {pref}{cmd.name.ljust(len_max)}  {cmd.short_doc}"

            runnables_hors_cog = await filter_runnables(cmd for cmd in ctx.bot.commands if not cmd.cog)
            if runnables_hors_cog:
                r += f"\n\nCommandes isolées :"
                for cmd in runnables_hors_cog:
                    r += f"\n  - {pref}{cmd.name.ljust(len_max)}  {cmd.short_doc}"

            r += f"\n\nUtilise <{pref}help command> pour plus d'information sur une commande."

            ctx.bot.in_command.append(ctx.channel.id)   # On réactive la limitation

        else:       # Aide détaillée sur une commande
            if command.startswith(pref):        # Si le joueur fait !help !command
                command = command.lstrip(pref)
            if command in aliases:              # Si !help d'un alias
                command = aliases[command]      # On remplace l'alias par sa commande

            if command in commandes:             # Si commande existante
                cmd = commandes[command]

                doc = cmd.help or ""
                doc = doc.replace("``", "`")
                doc = doc.replace("Args:", "Arguments :")
                doc = re.sub(r":\w+?:`[\.~!]*(.+?)`", r"`\1`", doc)

                r = f"{pref}{command} {cmd.signature} – {doc}\n"
                if cmd_aliases := [alias for alias,cmd in aliases.items() if cmd == command]:       # Si la commande a des alias
                    r += f"\nAlias : {pref}" + f", {pref}".join(cmd_aliases)

            else:
                r = f"Commande <{command}> non trouvée.\nUtilise <{pref}help> pour la liste des commandes."

        r += "\nSi besoin, n'hésite pas à appeler un MJ en les mentionnant (@MJ)."
        await tools.send_code_blocs(ctx, r, sep="\n\n")     # On envoie, en séparant en blocs de 2000 caractères max



# Définition classe principale

class LGBot(commands.Bot):
    """Bot Discord pour parties de Loup-Garou à la PCéenne.

    Classe fille de :class:`discord.ext.commands.Bot`, utilisable exactement de la même manière.
    """
    def __init__(self, command_prefix="!", description=None, case_insensitive=True,
                 intents=discord.Intents.all(), member_cache_flags=discord.MemberCacheFlags.all(), **kwargs):
        """Initialize self"""
        if not description:
            description = "LG-bot – Plateforme pour parties endiablées de Loup-Garou"

        super().__init__(
            command_prefix=command_prefix,
            description=description,
            case_insensitive=case_insensitive,
            intents=intents,
            member_cache_flags=member_cache_flags,
            **kwargs
        )

        #: :class:`int`: L'ID du serveur sur lequel tourne le bot.
        #: Vaut ``None`` avant l'appel à :meth:`run`, puis la valeur de la variable d'environnement ``LGREZ_SERVER_ID``.
        self.GUILD_ID = None

        #: :class:`dict`: Personalisation du bot (à définir avant l'appel à :meth:`run`), utilisable par les différentes fonctionnalités
        #: (exemples : ``bot.config["debut_saison"]``, ``bot.config["demande_porte"]``...)
        self.config = {}

        #: :class:`list`\[:class:`int`\]: IDs des salons dans lequels une commande est en cours d'exécution.
        self.in_command = []
        #: :class:`list`\[:class:`int`\]: IDs des salons en mode STFU.
        self.in_stfu = []
        #: :class:`list`\[:class:`int`\]: IDs des salons en mode Foire à la saucisse.
        self.in_fals = []
        #: :class:`dict`\[:class:`int` (:attr:`.bdd.Taches.id`), :class:`asyncio.TimerHandle`]): Tâches planifiées actuellement en attente.
        self.tasks = {}

        # Checks et système de blocage
        self.add_check(already_in_command)

        self.before_invoke(add_to_in_command)
        self.after_invoke(remove_from_in_command)

        # Chargement des commandes (définies dans les fichiers annexes, un cog par fichier dans features)
        self.add_cog(informations.Informations(self))            # Information du joueur
        self.add_cog(voter_agir.VoterAgir(self))                 # Voter ou agir
        self.add_cog(actions_publiques.ActionsPubliques(self))   # Haros et candidatures
        self.add_cog(IA.GestionIA(self))                         # Ajout, liste, modification des règles d'IA
        self.add_cog(open_close.OpenClose(self))                 # Ouverture/fermeture votes/actions (appel par webhook)
        self.add_cog(sync.Sync(self))                            # Synchronisation depuis GSheets
        self.add_cog(taches.GestionTaches(self))                 # Tâches planifiées
        self.add_cog(communication.Communication(self))          # Envoi de messages et embeds
        self.add_cog(annexe.Annexe(self))                        # Ouils divers et plus ou moins inutiles

        self.remove_command("help")
        self.add_cog(Special(self))         # Commandes spéciales, méta-commandes...


    # Réactions aux différents évènements
    async def on_ready(self):
        await _on_ready(self)
    on_ready.__doc__ = _on_ready.__doc__

    async def on_member_join(self, member):
        await _on_member_join(self, member)
    on_member_join.__doc__ = _on_member_join.__doc__

    async def on_member_remove(self, member):
        await _on_member_remove(self, member)
    on_member_remove.__doc__ = _on_member_remove.__doc__

    async def on_message(self, message):
        await _on_message(self, message)
    on_message.__doc__ = _on_message.__doc__

    async def on_raw_reaction_add(self, payload):
        await _on_raw_reaction_add(self, payload)
    on_raw_reaction_add.__doc__ = _on_raw_reaction_add.__doc__


    # Gestion des erreurs
    async def on_command_error(self, ctx, exc):
        await _on_command_error(self, ctx, exc)
    on_command_error.__doc__ = _on_command_error.__doc__

    async def on_error(self, event, *args, **kwargs):
        await _on_error(self, event, *args, **kwargs)
    on_error.__doc__ = _on_error.__doc__


    # Système de vérification de vie
    def i_am_alive(self, filename="alive.log"):
        """Exporte le temps actuel (UTC) et planifie un nouvel appel dans 60s

        Args:
            filename (:class:`str`): fichier où exporter le temps actuel
        """
        with open(filename, "w") as f:
            f.write(str(time.time()))
        self.loop.call_later(60, self.i_am_alive, filename)


    # Lancement du bot
    def run(self, *args, **kwargs):
        """Prépare puis lance le bot (bloquant).

        Récupère les informations de connection, établit la connection à la base de données puis lance le bot.

        Args:
            *args, \**kwargs: Passés à :meth:`discord.ext.commands.Bot.run`.
        """
        # Récupération du token du bot et de l'ID du serveur
        LGREZ_DISCORD_TOKEN = env.load("LGREZ_DISCORD_TOKEN")
        self.GUILD_ID = int(env.load("LGREZ_SERVER_ID"))

        # Connection BDD
        bdd.connect()

        # Lancement du bot
        super().run(LGREZ_DISCORD_TOKEN, *args, **kwargs)



"""lg-rez / blocs / Envoi de webhooks

(Implémentation de https://pypi.org/project/discord-webhook)

"""

import time

from discord_webhook import DiscordWebhook



def send(message, url):
    """Envoie un webhook Discord

    Args:
        message (:class:`str`): message à envoyer
        url (:class:`str`): adresse du webhook (``"https://discordapp.com/api/webhooks/.../..."``)

    Returns:
        :class:`requests.Response` | ``False``

    Seuls les webhooks textuels sont pris en charge.
    """
    webhook = DiscordWebhook(url=url, content=message)

    response = webhook.execute()

    return response



"""lg-rez / blocs / Lecture des variables d'environnement

"""

import sys
import os

from dotenv import load_dotenv


# À l'import
if len(sys.argv) > 1 and sys.argv[1].startswith(".env"):
    path = sys.argv[1]
else:
    path = ".env"

load_dotenv(path)


def load(VAR_NAME):
    """Lit une variable depuis les variables d'environnement / le ``.env`` demandé à l'appel de bot.py (``.env`` par défaut)

    Équivaut globalement à :func:`os.getenv` suivi d'une assertion vérifiant que la variable existe.

    Args:
        VAR_NAME (:class:`str`): nom de la variable à charger (``LGREZ_...``)

    Returns:
        :class:`str`
    """
    VAR = os.getenv(VAR_NAME)
    assert VAR, f"Variable d'environnement {VAR_NAME} manquante"
    return VAR



"""lg-rez / features / Synchronisation GSheets

Récupération et application des données des GSheets : modifications décidées via le Tableau de bord et rôles

"""

import traceback
import time

from discord import Embed
from discord.ext import commands

from lgrez.blocs import tools, bdd, bdd_tools, env, gsheets
from lgrez.blocs.bdd import engine, Tables, Joueurs, Actions, BaseActions, BaseActionsRoles, Roles, Taches
from lgrez.features import gestion_actions


class TDBModif():
    """Modification flag sur le Tableau de bord, à appliquer"""

    def __init__(self, id, col, val, row, column):
        """Initializes self."""
        #: :class:`int`: ID Discord du joueur concerné
        self.id = id
        #: :class:`str`: Colonne de Joueurs à modifier
        self.col = col
        #: :class:`object`: Nouvelle valeur
        self.val = val

        #: :class:`int`: Numéro de la ligne sur le TDB
        self.row = row
        #: :class:`int`: Numéro de la colonne TAMPON sur le TDB
        self.column = column

    def __repr__(self):
        """Returns repr(self)"""
        return f"<TDBModif id {self.id}: {self.col} = {self.val}>"


def get_sync():
    """Récupère les modifications en attente sur le TDB

    Charge les données du Tableau de bord (variable d'environment ``LGREZ_TDB_SHEET_ID``), compare les informations qui y figurent avec celles de la base de données (:class:`.bdd.Joueurs`)

    Returns:
        :class:`list`\[:class:`.TDBModif`\]: liste des modifications à apporter
    """

    all_cols = bdd_tools.get_cols(Joueurs)
    cols = [col for col in all_cols if not col.endswith('_')]    # On élimine les colonnes locales
    cols_SQL_types = bdd_tools.get_SQL_types(Joueurs)
    cols_SQL_nullable = bdd_tools.get_SQL_nullable(Joueurs)
    primary_col = bdd_tools.get_primary_col(Joueurs)

    ### RÉCUPÉRATION INFOS GSHEET

    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")

    workbook = gsheets.connect(SHEET_ID)    # Tableau de bord
    sheet = workbook.worksheet("Journée en cours")
    values = sheet.get_all_values()         # Liste de liste des valeurs des cellules

    head = values[2]            # Ligne d'en-têtes (noms des colonnes) = 3e ligne du TDB
    TDB_index = {col: head.index(col) for col in cols}    # Dictionnaire des indices des colonnes GSheet pour chaque colonne de la table
    TDB_tampon_index = {col: head.index(f"tampon_{col}") for col in cols if col != 'discord_id'}    # Idem pour la partie « tampon »

    # CONVERSION INFOS GSHEET EN UTILISATEURS

    joueurs_TDB = []            # Liste des joueurs tels qu'actuellement dans le TDB
    ids_TDB = []                # discord_ids des différents joueurs du TDB
    rows_TDB = {}               # Indices des lignes ou sont les différents joueurs du TDB

    for i_row in range(len(values)):
        L = values[i_row]           # On parcourt les lignes du TDB
        id_cell = L[TDB_index[primary_col]]
        if id_cell.isdigit():        # Si la cellule contient bien un ID (que des chiffres, et pas vide)
            id = int(id_cell)
            joueur_TDB = {col: bdd_tools.transtype(L[TDB_index[col]], col, cols_SQL_types[col], cols_SQL_nullable[col]) for col in cols}    # Dictionnaire correspondant à l'utilisateur
            joueurs_TDB.append(joueur_TDB)
            ids_TDB.append(id)
            rows_TDB[id] = i_row

    ### RÉCUPÉRATION UTILISATEURS CACHE

    joueurs_BDD = Joueurs.query.all()     # Liste des joueurs tels qu'actuellement en cache
    ids_BDD = [joueur_BDD.discord_id for joueur_BDD in joueurs_BDD]

    ### COMPARAISON

    modifs = []         # modifs à porter au TDB (liste de TDBModifs)

    for joueur_BDD in joueurs_BDD.copy():           # Joueurs dans le cache supprimés du TDB
        if joueur_BDD.discord_id not in ids_TDB:
            joueurs_BDD.remove(joueur_BDD)
            bdd.session.delete(joueur_BDD)

    for joueur_TDB in joueurs_TDB:                  # Différences
        id = joueur_TDB["discord_id"]

        if id not in ids_BDD:             # Si joueur dans le cache pas dans le TDB
            raise ValueError(f"Joueur {joueur_TDB['nom']} hors BDD : vérifier processus d'inscription")

        joueur_BDD = [joueur for joueur in joueurs_BDD if joueur.discord_id == id][0]     # joueur correspondant dans le cache

        for col in cols:
            if getattr(joueur_BDD, col) != joueur_TDB[col]:     # Si <col> diffère entre TDB et cache
                modif = TDBModif(id, col, joueur_TDB[col],
                                 rows_TDB[id], TDB_tampon_index[col])
                modifs.append(modif)   # On ajoute les modifs

    ### RETOURNAGE DES RÉSULTATS

    return modifs


def validate_sync(modifs):
    """Valide des modificatons sur le Tableau de bord (case plus en rouge)

    Args:
        modifs (:class:`list`\[:class:`.TDBModif`\]): liste des modifications à apporter

    Modifie sur le Tableau de bord (variable d'environment ``LGREZ_TDB_SHEET_ID``) et applique les modifications contenues dans modifs
    """

    ### APPLICATION DES MODIFICATIONS SUR LE TDB

    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")

    workbook = gsheets.connect(SHEET_ID)    # Tableau de bord
    sheet = workbook.worksheet("Journée en cours")

    modifs_lc = [(modif.row, modif.column, modif.val) for modif in modifs]
        # On transforme les infos en coordonnées dans le TDB : ID -> ligne et col -> colonne,
    gsheets.update(sheet, modifs_lc)


async def modif_joueur(ctx, joueur_id, modifs, silent=False):
    """Attribue les modifications demandées au joueur

    Args:
        ctx (:class:`~discord.ext.commands.Context`): contexte quelconque du bot
        joueur_id (:class:`int`): id Discord du joueur concerné
        modifs (:class:`list`\[:class:`.TDBModif`\]): liste des modifications à apporter
        silent (:class:`bool`): si ``True``, no notifie pas le joueur des modifications

    Pour chaque modifications de ``modif``, applique les conséquences adéquates (rôles, nouvelles actions, tâches planifiées...) et informe le joueur si ``silent`` vaut ``False``.
    """
    joueur = Joueurs.query.get(int(joueur_id))
    assert joueur, f"!sync : joueur d'ID {joueur_id} introuvable"

    member = ctx.guild.get_member(joueur.discord_id)
    assert member, f"!sync : member {joueur} introuvable"

    chan = ctx.guild.get_channel(joueur.chan_id_)
    assert chan, f"!sync : chan privé de {member} introuvable"

    changelog = f"\n- {member.display_name} (@{member.name}#{member.discriminator}) :\n"
    notif = ""

    for modif in modifs:
        changelog += f"    - {modif.col} : {modif.val}\n"

        if modif.col == "nom":                            # Renommage joueur
            await chan.edit(name=f"conv-bot-{modif.val}")
            await member.edit(nick=modif.val)
            if not silent:
                notif += f":arrow_forward: Tu t'appelles maintenant {tools.bold(modif.val)}.\n"

        elif modif.col == "chambre" and not silent:       # Modification chambre
            notif += f":arrow_forward: Tu habites maintenant en chambre {tools.bold(modif.val)}.\n"

        elif modif.col == "statut":
            if modif.val == "vivant":                     # Statut = vivant
                await member.add_roles(tools.role(ctx, "Joueur en vie"))
                await member.remove_roles(tools.role(ctx, "Joueur mort"))
                if not silent:
                    notif += f":arrow_forward: Tu es maintenant en vie. EN VIE !!!\n"

            elif modif.val == "mort":                     # Statut = mort
                await member.add_roles(tools.role(ctx, "Joueur mort"))
                await member.remove_roles(tools.role(ctx, "Joueur en vie"))
                if not silent:
                    notif += f":arrow_forward: Tu es malheureusement décédé(e) :cry:\nÇa arrive même aux meilleurs, en espérant que ta mort ait été belle !\n"
                # Actions à la mort
                for action in Actions.query.filter_by(player_id=joueur.discord_id, trigger_debut="mort"):
                    await gestion_actions.open_action(ctx, action, chan)

            elif modif.val == "MV":                       # Statut = MV
                await member.add_roles(tools.role(ctx, "Joueur en vie"))
                await member.remove_roles(tools.role(ctx, "Joueur mort"))
                if not silent:
                    notif += f":arrow_forward: Te voilà maintenant réduit(e) au statut de mort-vivant... Un MJ viendra te voir très vite, si ce n'est déjà fait, mais retient que la partie n'est pas finie pour toi !\n"

            elif not silent:                        # Statut = autre
                notif += f":arrow_forward: Nouveau statut : {tools.bold(modif.val)} !\n"

        elif modif.col == "role":                         # Modification rôle
            old_bars = BaseActionsRoles.query.filter_by(role=joueur.role).all()
            old_actions = []
            for bar in old_bars:
                old_actions.extend(Actions.query.filter_by(action=bar.action, player_id=joueur.discord_id).all())
            for action in old_actions:
                gestion_actions.delete_action(ctx, action)  # On supprime les anciennes actions de rôle (et les tâches si il y en a)

            new_bars = BaseActionsRoles.query.filter_by(role=modif.val).all()         # Actions associées au nouveau rôle
            new_bas = [BaseActions.query.get(bar.action) for bar in new_bars]   # Nouvelles BaseActions
            cols = [col for col in bdd_tools.get_cols(BaseActions) if not col.startswith("base")]
            new_actions = [Actions(player_id=joueur.discord_id, **{col: getattr(ba, col) for col in cols},
                                   cooldown=0, charges=ba.base_charges) for ba in new_bas]
            await tools.log(ctx, str(new_actions))

            for action in new_actions:
                gestion_actions.add_action(ctx, action)     # Ajout et création des tâches si trigger temporel

            role = tools.nom_role(modif.val)
            if not role:        # role <modif.val> pas en base : Error!
                role = f"« {modif.val} »"
                await tools.log(ctx, f"{tools.mention_MJ(ctx)} ALED : rôle \"{modif.val}\" attribué à {joueur.nom} inconnu en base !")
            if not silent:
                notif += f":arrow_forward: Ton nouveau rôle, si tu l'acceptes : {tools.bold(role)} !\nQue ce soit pour un jour ou pour le reste de la partie, renseigne toi en tapant {tools.code(f'!roles {modif.val}')}.\n"

        elif modif.col == "camp" and not silent:          # Modification camp
            notif += f":arrow_forward: Tu fais maintenant partie du camp « {tools.bold(modif.val)} ».\n"

        elif modif.col == "votant_village" and not silent:
            if modif.val:                                 # votant_village = True
                notif += f":arrow_forward: Tu peux maintenant participer aux votes du village !\n"
            else:                                   # votant_village = False
                notif += f":arrow_forward: Tu ne peux maintenant plus participer aux votes du village.\n"

        elif modif.col == "votant_loups" and not silent:
            if modif.val:                                 # votant_loups = True
                notif += f":arrow_forward: Tu peux maintenant participer aux votes des loups ! Amuse-toi bien :wolf:\n"
            else:                                   # votant_loups = False
                notif += f":arrow_forward: Tu ne peux maintenant plus participer aux votes des loups.\n"

        elif modif.col == "role_actif" and not silent:
            if modif.val:                                 # role_actif = True
                notif += f":arrow_forward: Tu peux maintenant utiliser tes pouvoirs !\n"
            else:                                   # role_actif = False
                notif += f":arrow_forward: Tu ne peux maintenant plus utiliser aucun pouvoir.\n"

        bdd_tools.modif(joueur, modif.col, modif.val)           # Dans tous les cas, on modifie en base (après, pour pouvoir accéder aux vieux attribus plus haut)

    if not silent:
        await chan.send(f":zap: {member.mention} Une action divine vient de modifier ton existence ! :zap:\n"
                        + f"\n{notif}\n"
                        + tools.ital(":warning: Si tu penses qu'il y a erreur, appelle un MJ au plus vite !"))

    return changelog


class Sync(commands.Cog):
    """Sync - Commandes de synchronisation des GSheets vers la BDD et les joueurs"""

    @commands.command()
    @tools.mjs_only
    async def sync(self, ctx, silent=False):
        """Récupère et applique les modifications du Tableau de bord (COMMANDE MJ)

        Args:
            silent: si spécifié (quelque soit sa valeur), les joueurs ne sont pas notifiés des modifications.

        Cette commande va récupérer les modifications en attente sur le Tableau de bord (lignes en rouge), modifer la BDD Joueurs, et appliquer les modificatons dans Discord le cas échéant : renommage des utilisateurs, modification des rôles...
        """
        await ctx.send("Récupération des modifications...")
        async with ctx.typing():
            modifs = get_sync()                 # Récupération du dictionnaire {joueur_id: modified_attrs}
            silent = bool(silent)
            changelog = f"Synchronisation TDB (silencieux = {silent}) :"

        if not modifs:
            await ctx.send("Pas de nouvelles modificatons.")
            return

        dic = {}
        for modif in modifs:
            if modif.id not in dic:
                dic[modif.id] = []
            dic[modif.id].append(modif)

        message = await ctx.send(f"{len(modifs)} modification(s) trouvée(s) pour {len(dic)} joueur(s), go ?")
        if not await tools.yes_no(ctx.bot, message):
            await ctx.send("Mission aborted.")
            return

        # Go sync
        done = []
        async with ctx.typing():
            for joueur_id, modifs in dic.items():        # Joueurs dont au moins un attribut a été modifié
                try:
                    changelog += await modif_joueur(ctx, joueur_id, modifs, silent)
                except Exception:
                    changelog += traceback.format_exc()
                    await ctx.send(f"Erreur joueur {joueur_id}, passage au suivant, voir logs pour les détails")
                else:
                    done.extend(modifs)

            bdd.session.commit()

            if done:
                validate_sync(done)

            await tools.log(ctx, changelog, code=True)

        await ctx.send(f"Fait (voir {tools.channel(ctx, 'logs').mention} pour le détail)")


    @commands.command()
    @tools.mjs_only
    async def fillroles(self, ctx):
        """Remplit les tables des rôles / actions et #roles depuis le GSheet ad hoc (COMMANDE MJ)

        - Remplit les tables :class:`.bdd.Roles`, :class:`.bdd.BaseActions` et :class:`.bdd.BaseActionsRoles` avec les informations du Google Sheets "Rôles et actions" (variable d'environnement ``LGREZ_ROLES_SHEET_ID``) ;
        - Vide le chan ``#roles`` puis le remplit avec les descriptifs de chaque rôle.

        Utile à chaque début de saison / changement dans les rôles/actions. Écrase toutes les entrées déjà en base, mais ne supprime pas celles obsolètes.
        """

        SHEET_ID = env.load("LGREZ_ROLES_SHEET_ID")
        workbook = gsheets.connect(SHEET_ID)    # Tableau de bord

        for table_name in ["Roles", "BaseActions", "BaseActionsRoles"]:
            await ctx.send(f"Remplissage de la table {tools.code(table_name)}...")
            async with ctx.typing():

                sheet = workbook.worksheet(table_name)
                values = sheet.get_all_values()         # Liste de liste des valeurs des cellules

                table = Tables[table_name]
                cols = bdd_tools.get_cols(table)
                SQL_types = bdd_tools.get_SQL_types(table)
                SQL_nullable = bdd_tools.get_SQL_nullable(table)
                primary_col = bdd_tools.get_primary_col(table)

                cols_index = {col: values[0].index(col) for col in cols}    # Dictionnaire des indices des colonnes GSheet pour chaque colonne de la table

                existants = {getattr(item, primary_col):item for item in table.query.all()}

                for L in values[1:]:
                    args = {col: bdd_tools.transtype(L[cols_index[col]], col, SQL_types[col], SQL_nullable[col]) for col in cols}
                    id = args[primary_col]
                    if id in existants:
                        for col in cols:
                            if getattr(existants[id], col) != args[col]:
                                bdd_tools.modif(existants[id], col, args[col])
                    else:
                        bdd.session.add(table(**args))

                bdd.session.commit()

            await ctx.send(f"Table {tools.code(table_name)} remplie !")
            await tools.log(ctx, f"Table {tools.code(table_name)} remplie !")

        chan_roles = tools.channel(ctx, "rôles")

        await ctx.send(f"Vidage de {chan_roles.mention}...")
        async with ctx.typing():
            await chan_roles.purge(limit=1000)

        roles = {camp: Roles.query.filter_by(camp=camp).all() for camp in ["village", "loups", "nécro", "solitaire", "autre"]}
        await ctx.send(f"Remplissage... (temps estimé : {sum([len(v) + 2 for v in roles.values()]) + 1} secondes)")

        t0 = time.time()
        await chan_roles.send(f"Voici la liste des rôles : (accessible en faisant {tools.code('!roles')}, mais on l'a mis là parce que pourquoi pas)\n\n——————————————————————————")
        async with ctx.typing():
            for camp, roles_camp in roles.items():
                if roles_camp:
                    await chan_roles.send(embed=Embed(title=f"Camp : {camp}").set_image(url=tools.emoji_camp(ctx, camp).url))
                    await chan_roles.send(f"——————————————————————————")
                    for role in roles_camp:
                        await chan_roles.send(f"{tools.emoji_camp(ctx, role.camp)} {tools.bold(role.prefixe + role.nom)} – {role.description_courte} (camp : {role.camp})\n\n{role.description_longue}\n\n——————————————————————————")

        await ctx.send(f"{chan_roles.mention} rempli ! (en {(time.time() - t0):.4} secondes)")



"""lg-rez / blocs / Gestion des données

Déclaration de toutes les tables et leurs colonnes, et connection à la BDD

"""

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
import psycopg2

from lgrez.blocs import env


#: Alias de :exc:`sqlalchemy.exc.SQLAlchemyError` : exception de BDD générale
SQLAlchemyError = sqlalchemy.exc.SQLAlchemyError

#: Alias de :exc:`psycopg2.SQLAlchemyError` : erreur levée en cas de perte de connection avec la BDD
#: Seul PostreSQL géré nativement, override DriverOperationalError avec l'équivalent pour un autre driver
DriverOperationalError = psycopg2.OperationalError


Base = declarative_base()
Base.__doc__ = "Classe de base des tables de données, renvoyée par :func:`sqlalchemy.ext.declarative.declarative_base`"
#: :class:`dict`\[:class:`str`, :class:`.Base`\]: Dictionnaire {nom de la base -> table}
Tables = {}


# Objets de connection (créés dans connect)
#: :class:`sqlalchemy.engine.Engine`: Moteur de connection à la BDD
#: Vaut ``None`` avant l'appel à :func:`connect`.
engine = None
#: :class:`sqlalchemy.orm.session.Session`: Session de transaction avec la BDD.
#: Vaut ``None`` avant l'appel à :func:`connect`.
session = None


class _ClassProperty(property):      # https://stackoverflow.com/a/1383402
    """Permet d'utiliser des propriés comme des classmethods (reçoivent la classe comme seul argument)"""
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class _MyTable(object):
    """Classe permettant d'enregister les tables au fur et à mesure de leur déclaration et d'utiliser Table.query au lieu de session.query(Table)"""
    def __init_subclass__(cls, **kwargs):
        """Définit __tablename__ et enregistre la table dans Tables"""
        super().__init_subclass__(**kwargs)
        cls.__tablename__ = cls.__name__.lower()
        Tables[cls.__name__] = cls
        cls.__doc__ += f"\n\nClasse de données (sous-classe de :class:`.Base`) représentant la table ``{cls.__tablename__}``.\n\nTous les attributs ci-dessous sont du type indiqué pour les instances (entrées de BDD), mais de type :class:`sqlalchemy.orm.attributes.InstrumentedAttribute` pour la classe elle-même."       # On adore la doc vraiment

    @classmethod
    def get_query(cls):
        return session.query(cls)

    query = _ClassProperty(get_query)        # Permet d'utiliser Table.query.<...> au lieu de bdd.session.query(Table).<...>



# Définition des tables

class Joueurs(_MyTable, Base): #, tablename="joueurs"
    """Table de données des joueurs inscrits

    Les instances de cette classe correspondent aux lignes du Tableau de bord ; elles sont crées par l'inscription (:func:`.inscription.main`) et synchronisées par :meth:`\!sync <.sync.Sync.Sync.sync.callback>`.
    """
    #: :class:`int` : ID Discord du joueur (clé primaire)
    discord_id = sqlalchemy.Column(sqlalchemy.BigInteger(), primary_key=True)
    #: :class:`int`: ID du chan privé Discord du joueur (NOT NULL)
    #: Le ``_`` final indique que ce champ n'est pas synchnisé avec le Tableau de bord.
    chan_id_ = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)

    #: :class:`str`: nom du joueur (demandé à l'inscription) (NOT NULL, ``len <= 32``)
    nom = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    #: :class:`str`: emplacement du joueur (demandé à l'inscription) (``len <= 200``)
    chambre = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)
    #: :class:`str`: statut RP (généralement "vivant", "mort" ou "MV") (NOT NULL, ``len <= 32``)
    statut = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    #: :class:`str`: rôle du joueur : doit correspondre à une valeur de :attr:`Roles.slug` (NOT NULL, ``len <= 32``)
    role = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    #: :class:`str`: camp du joueur (généralement "village", "loups", "nécro") (NOT NULL, ``len <= 32``)
    camp = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    #: :class:`bool`: si le joueur participe aux votes du village ou non
    votant_village = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=False)
    #: :class:`bool`: si le joueur participe au vote des loups ou non
    votant_loups = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=False)
    #: :class:`bool`: si le peut agir ou non (chatgarouté...) (NOT NULL)
    role_actif = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    #: :class:`str`: vote actuel au vote condamné (None si pas de vote en cours) (``len <= 200``)
    #: Le ``_`` final indique que ce champ n'est pas synchnisé avec le Tableau de bord.
    vote_condamne_ = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)
    #: :class:`str`: vote actuel au vote maire (None si pas de vote en cours) (``len <= 200``)
    #: Le ``_`` final indique que ce champ n'est pas synchnisé avec le Tableau de bord.
    vote_maire_ = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)
    #: :class:`str`: vote actuel au vote loups (None si pas de vote en cours) (``len <= 200``)
    #: Le ``_`` final indique que ce champ n'est pas synchnisé avec le Tableau de bord.
    vote_loups_ = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)

    def __repr__(self):
        """Return repr(self)."""
        return f"<Joueurs #{self.discord_id} ({self.nom})>"



class Roles(_MyTable, Base): #, tablename="roles"
    """Table de données des rôles

    Cette table est remplie automatiquement à partir du Google Sheet "Rôles et actions" par la commande :meth:`\!fillroles <.sync.Sync.Sync.fillroles.callback>`.
    """

    #: :class:`str`: identifiant unique du rôle (clé primaire, ``len <= 32``)
    slug = sqlalchemy.Column(sqlalchemy.String(32), primary_key=True)

    #: :class:`str`: article du nom du rôle (``"Le"``, ``"La"``, ``"L'"``...) (NOT NULL, ``len <= 8``)
    prefixe = sqlalchemy.Column(sqlalchemy.String(8), nullable=False)
    #: :class:`str`: nom du rôle (NOT NULL, ``len <= 32``)
    nom = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    #: :class:`str`: camp de base du rôle (NOT NULL, ``len <= 32``)
    camp = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)                       # loups, solitaire, nécro, village...

    #: :class:`str`: description en une ligne (NOT NULL, ``len <= 140``)
    description_courte = sqlalchemy.Column(sqlalchemy.String(140), nullable=False)
    #: :class:`str`: règles et background complets (NOT NULL, ``len <= 2000``)
    description_longue = sqlalchemy.Column(sqlalchemy.String(2000), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<Roles '{self.slug}' ({self.prefixe}{self.nom})>"



class BaseActions(_MyTable, Base): #, tablename="base_actions"
    """Table de données des actions définies de bases (non liées à un joueur)

    Cette table est remplie automatiquement à partir du Google Sheet "Rôles et actions" par la commande :meth:`\!fillroles <.remplissage_bdd.RemplissageBDD.RemplissageBDD.fillroles.callback>`.
    """
    #: :class:`str`: slug identifiant uniquement l'action (primary key, ``len <= 2000``)
    action = sqlalchemy.Column(sqlalchemy.String(32), primary_key=True)

    #: :class:`str`: type de déclencheur du début de l'action (``"temporel"``, ``"mort"``, ...) (NOT NULL, ``len <= 2000``)
    trigger_debut = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    #: :class:`str`: type de déclencheur de la fin (NOT NULL, ``len <= 2000``)
    trigger_fin = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    #: :class:`bool`: si l'action est instantannée (conséquences dès la prise de décision) ou non (conséquence à la fin du créneau d'action)
    instant = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    #: :class:`datetime.time`: si :attr:`.trigger_debut` vaut ``"temporel"`` ou ``"delta"``, l'horaire / la durée associée
    heure_debut = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)
    #: :class:`datetime.time`: si :attr:`.trigger_fin` vaut ``"temporel"`` ou ``"delta"``, l'horaire / la durée associée
    heure_fin = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)

    #: :class:`int`: temps de rechargement entre deux utilisations du pouvoir (``0`` si pas de cooldown) (NOT NULL)
    base_cooldown = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)
    #: :class:`int`: nombre de charges initiales du pouvoir (``None`` si illimitée)
    base_charges = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)
    #: :class:`str`: évènements pouvant recharger le pouvoir, séparés par des virgules (parmi ``"weekends"``, ``"forgeron"``, ...) (NOT NULL, ``len <= 32``)
    refill  = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)

    #: :class:`str`: attribut informatif (Distance/Physique/Lieu/Contact/Conditionnel/None/Public) (``len <= 32``)
    lieu = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    #: :class:`str`: attribut informatif (Oui, Non, Conditionnel, Potion, Rapport; None si récursif) (``len <= 32``)
    interaction_notaire = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    #: :class:`str`: attribut informatif (Oui, Non, Conditionnel, Taverne, Feu, MaisonClose, Précis, Cimetière, Loups, None si recursif) (``len <= 32``)
    interaction_gardien = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    #: :class:`str`: attribut informatif (Oui, Non, changement de cible, etc) (``len <= 100`` 32)
    mage = sqlalchemy.Column(sqlalchemy.String(100), nullable=True)

    #: :class:`bool`: si la cible doit changer entre deux utilisations consécutives (informatif uniquement pour l'instant)
    changement_cible = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseActions '{self.action}'>"



class Actions(_MyTable, Base): #, tablename="actions"
    """Table de données des actions attribuées (liées à un joueur et actives)

    Les instances doivent être enregistrées via :func:`.gestion_actions.open_action` et supprimées via :func:`.gestion_actions.close_action`.
    """
    #: :class:`int`: identifiant unique de l'action, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    #: :class:`int`: clé étrangère liée à :attr:`.Joueurs.id` (NOT NULL)
    player_id = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)
    #: :class:`str`: clé étrangère liée à :attr:`.BaseActions.slug` (NOT NULL, ``len <= 32``)
    action = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    #: et ``trigger_fin``, ``instant``, ``heure_debut``, ``heure_fin``, ``instant``, ``heure_debut``, ``heure_fin``, ``cooldown``, ``charges``, ``refill``, ``lieu``, ``interaction_notaire``, ``interaction_gardien``, ``mage``, ``changement_cible`` sont les mêmes que pour :class:`.Actions`.
    trigger_debut = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    trigger_fin = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    instant = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    heure_debut = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)
    heure_fin = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)

    cooldown = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)
    charges = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)
    refill  = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)

    lieu = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    interaction_notaire = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    interaction_gardien = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    mage = sqlalchemy.Column(sqlalchemy.String(100), nullable=True)
    changement_cible = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    #: Décision prise par le joueur pour l'action actuelle (``None`` si action non en cours). Le ``_`` final indique que ce champ n'est pas synchnisé avec le Tableau de bord. (``len <= 200``)
    decision_ = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<Actions #{self.id} ({self.action}/{self.player_id})>"



class BaseActionsRoles(_MyTable, Base): #, tablename="base_action_roles"
    """Table de données mettant en relation les rôles et les actions de base

    Cette table est remplie automatiquement à partir du Google Sheet "Rôles et actions" par la commande :meth:`\!fillroles <.remplissage_bdd.RemplissageBDD.RemplissageBDD.fillroles.callback>`.
    """
    #: :class:`int`: identifiant unique de l'action, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    #: :class:`str`: clé étrangère liée à :attr:`.Roles.slug` (NOT NULL, ``len <= 32``)
    role = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    #: :class:`str`: clé étrangère liée à :attr:`.BaseActions.slug` (NOT NULL, ``len <= 32``)
    action = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseActionsRoles #{self.id} ({self.role}/{self.action})>"



class Taches(_MyTable, Base): #, tablename="taches"
    """Table de données des tâches planifiées du bot

    Les instances doivent être enregistrées via :func:`.taches.add_task` et supprimées via :func:`.taches.cancel_task`.
    """
    #: :class:`int`: identifiant unique de la tâche, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    #: :class:`datetime.datetime`: moment où sera exécutée la tâche (NOT NULL)
    timestamp = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False)
    #: :class:`str`: texte à envoyer via le webhook (généralement une commande) (NOT NULL, ``len <= 200``)
    commande = sqlalchemy.Column(sqlalchemy.String(2000), nullable=False)
    #: :class:`int`: si la tâche est liée à une action, clé étrangère liée à :attr:`.Actions.id`
    action = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<Taches #{self.id} ({self.commande})>"



class Reactions(_MyTable, Base): #, tablename="reactions"
    """Table de données des réactions d'IA connues du bot

    Les instances doivent être enregistrées via :meth:`\!addIA <.IA.GestionIA.GestionIA.addIA.callback>` et supprimées via :meth:`\!modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """
    #: :class:`int`: identifiant unique de la réaction, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    #: :class:`str`: réponse, suivant le format (mini-langage) personnalisé (``"txt <||> txt <&&> <##>react"``) (NOT NULL, <2000)
    reponse = sqlalchemy.Column(sqlalchemy.String(2000), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        extract = self.reponse.replace('\n', ' ')[:15] + "..."
        return f"<Reactions #{self.id} ({extract})>"



class Triggers(_MyTable, Base): #, tablename="triggers"
    """Table de données des mots et expressions déclenchant l'IA du bot

    Les instances doivent être enregistrées via :meth:`\!addIA <.IA.GestionIA.GestionIA.addIA.callback>` et supprimées via :meth:`\!modifIA <.IA.GestionIA.GestionIA.modifIA.callback>`.
    """
    #: :class:`int`: identifiant unique du trigger, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    #: :class:`str`: mots-clés / expressions, séparés par des ``;`` éventuellement (NOT NULL, ``len <= 500``)
    trigger = sqlalchemy.Column(sqlalchemy.String(500), nullable=False)
    #: :class:`int`: réaction associée, clé étrangère de :attr:`.Reactions.id` (NOT NULL)
    reac_id = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)


    def __repr__(self):
        """Return repr(self)."""
        extract = self.trigger.replace('\n', ' ')[:15] + "..."
        return f"<Triggers #{self.id} ({extract})>"



class CandidHaro(_MyTable, Base): #, tablename="candid_haro"
    """Table de données des candidatures et haros en cours  #PhilippeCandidHaro

    Les instances doivent être enregistrées via :meth:`\!haro <.actions_publiques.ActionsPubliques.ActionsPubliques.haro.callback>` / :meth:`\!candid <.actions_publiques.ActionsPubliques.ActionsPubliques.candid.callback>` et supprimées via :meth:`\!wipe <.actions_publiques.ActionsPubliques.ActionsPubliques.wipe.callback>`.
    """
    #: :class:`int`: identifiant unique de la réaction, sans signification (auto-incrémental) (primary key)
    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key = True)
    #: :class:`int`: joueur associé, clé étrangère de :attr:`.Joueurs.discord_id` (NOT NULL)
    player_id = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)
    #: :class:`str`: ``"haro"`` ou ``"candidature"`` (flemme de ``enum`` mais faudrait, vraiment) (NOT NULL, ``len <= 11``)
    type = sqlalchemy.Column(sqlalchemy.String(11), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<CandidHaro #{self.id} ({self.player_id}/{self.type})>"



def connect():
    """Se connecte à la base de données (variable d'environment ``LGREZ_DATABASE_URI``), crée les tables si nécessaire, construit :attr:`.bdd.engine` et ouvre :attr:`.bdd.session`"""
    global engine, session

    LGREZ_DATABASE_URI = env.load("LGREZ_DATABASE_URI")
    engine = sqlalchemy.create_engine(LGREZ_DATABASE_URI, pool_pre_ping=True)           # Moteur SQL : connexion avec le serveur

    # Création des tables si elles n'existent pas déjà
    Base.metadata.create_all(engine)

    # Ouverture de la session
    Session = sessionmaker(bind=engine)
    session = Session()
